"""Summoner Common Functions / Definitions"""
from enum import IntEnum
import os

import gitlab
import pypd

# pylint: disable=W0107

from .common_logger import Logger

LOGGER = Logger().get_logger(__name__)

class ErrorCodes(IntEnum):
    """
    A simple enum to make it easier to return particular error codes.
    Added specifically to aid in testing various stages of failure.
    """
    SUCCESS = 0                      # Yeah, a bit weird, but this "error" actually is no error. Successful run.
    ERROR_TERRAFORM_FORMAT = 100     # The terraform code failed a formatting check
    ERROR_BRANCH_DIVERGED = 105      # The current git branch is out of sync with master
    ERROR_TERRAFORM_INIT = 110       # The terraform code failed to initialize
    ERROR_TERRAFORM_VALIDATE = 115   # The terraform code failed a validation pass
    ERROR_TF_LINT = 120              # The terraform code failed a linting pass
    ERROR_TERRAFORM_PLAN = 125       # The terraform code failed to run plan
    ERROR_MISSING_PARAMS = 130       # The parameter required was not
    ERROR_MISSING_CONFIG = 135       # A configuration/environment config setting is missing
    ERROR_TERRAFORM_APPLY = 140      # The terraform code failed an apply pass
    ERROR_TERRAFORM_DESTROY = 145    # The terraform code failed a destroy pass
    ERROR_GITLAB = 135               # The Gitlab integration failed
    ERROR_LAUNCH_RECOVERY = 200      # WARN. A launch failure! Rollback successful though, things are stable.
    ERROR_LAUNCH_FAILURE = 210       # BAD. A launch failure! Something is broken!

class NotificationType(IntEnum):
    """
    A simple enum to help note which
    kind of notification we want to send.
    """
    LOW_PRIORITY = 5
    HIGH_PRIORITY = 10

class PagerDutyNotificationError(Exception):
    """
    PagerDuty notification Exception
    """
    pass

class DuplicateDeployCombination(ValueError):
    """
    This error is raised when a duplicate combination
    of Account Type, AWS Account, and Region is detected.
    For example, two of: (prod, 000, us-east-1).

    This is considered an error because we should only
    deploy once to such a combination.
    """
    pass

class TerraformInitFailure(Exception):
    """
    An exception raised if there is an issue running
    terraform init on a provided directory
    """
    pass

class DeployFailure(Exception):
    """
    This exception is raised when something in the
    deployment attempt failed.
    """
    pass

class RollbackComplete(DeployFailure):
    """
    This exception is raised when there is a failure
    in the deployment of an environment. However, in this
    case the rollback was successfully performed and the
    target is in a stable state.
    """
    pass

class RollbackFailure(DeployFailure):
    """
    This exception is raised when there is a failure
    in the rollback of an environment. This type of
    deployment failure is fatal and likely needs
    human intervention to correct
    """
    pass

class ConfigError(Exception):
    """
    This exception is raised when there is a failure
    importing values from the config. This type of
    failure is likely due to a missing value from
    the .toml file
    """
    pass

def set_severity_level(priority_level):
    """
    Set severity level of a PagerDuty notification.
    Arguments:
        priority_level -- (NotificationType) Priority level used to set severity.
    Returns:
        severity -- {info, critical}
    """
    if priority_level == NotificationType.LOW_PRIORITY:
        severity = "info"
    elif priority_level == NotificationType.HIGH_PRIORITY:
        severity = "critical"
    return severity

def send_pd_notification(integration_key, priority_level, source, summary):
    """
    Sends PagerDuty event notification .
    Arguments:
        integration_key -- (string) PagerDuty integration key used for routing events
        priority_level -- (NotificationType) Priority level used to set severity.
        message -- (string) A brief text summary of the event.
    """
    severity = set_severity_level(priority_level)

    notification_response = pypd.EventV2.create(data={
        'routing_key': integration_key,
        'event_action': 'trigger',
        'payload': {
            'summary': summary,
            'severity': severity,
            'source': 'Summoner: {}'.format(source)
        }
    })

    if not notification_response["status"] == "success":
        raise PagerDutyNotificationError(
            "PagerDuty notification event error. Recieved status: {}".format(notification_response["status"]))

def gen_terraform_backend_config(target_config, file_name, environment):
    """
    Generate a simple terraform backend file for the target account.

    Params:
      target_config: a config dictionary for the target account
      file_name: name of the backend file to write

    This function will read the config settings for the account from the provided
    target_config parameter to generate the remote state config for terraform.
    """

    # the terraform backend file is simply a flat file
    # with key=value pairs in it. We'll set the settings first.
    backend_s3 = dict()
    backend_s3["region"] = target_config['terraform']['state']['s3_region']
    backend_s3["bucket"] = target_config['terraform']['state']['s3_bucket']
    backend_s3["key"] = target_config['terraform']['state']['s3_path']
    backend_s3["dynamodb_table"] = target_config['terraform']['state']['dynamodb_table']
    if 's3_access_key' in target_config['terraform']['state'] and 's3_secret_key' in target_config['terraform']['state']:
        backend_s3["access_key"] = environment.get(target_config['terraform']['state']['s3_access_key'])
        backend_s3["secret_key"] = environment.get(target_config['terraform']['state']['s3_secret_key'])
    # With the settings, now we just output the contents to disk
    with open(file_name, 'w') as backend_file:
        for key, value in backend_s3.items():
            backend_file.write('{} = "{}"\n'.format(key, value))

def calculate_deploy_targets(config):
    """
    A helper function to loop over all aws accounts
    specified in the configuration and create a dictionary
    key'd by environment set to a list of configuration
    dictionaries.
      dict["test"] = [ {account_number: 0000, aws_access_key: "", etc.}, {}, {}.. ]
      dict["stage"] = [ {account_number: 0000, aws_access_key: "", etc.}, {}, {}.. ]

    Args:
      config:  the Summoner Configuration dictionary

    Returns:
      A dictionary as specified above.

    Raises:
      ValueError - if a duplicate (env, acct_num, region) combination exists.
                   we only want to deploy to a unique combination once.
    """

    # we're going to use this variable to ensure
    # that there is always a unique (env, acct_num, region) combination.
    # We don't want to deploy to the same thing twice.
    unique_combinations = set()

    # the targets to deploy to (will be returned to caller)
    deploy_targets = dict()

    # loop over each account in the config and based
    # on the account type specified, add it to the correct
    # deploy target list.
    for account in config['aws-account']:
        account_type = account["type"]
        account_number = account["account_number"]
        deploy_region = account["deploy_region"]

        # check that we have a unique combination here
        combination = (account_type, account_number, deploy_region) if "deploy_environment" not in account.keys() else (account_type, account_number, deploy_region, account["deploy_environment"])
        if combination in unique_combinations:
            raise DuplicateDeployCombination("Received a duplicate \
                type-number-region combination: " + str(combination))
        unique_combinations.add(combination)

        # insert our deploy targets for this type
        if account_type not in deploy_targets:
            deploy_targets[account_type] = list()
        deploy_targets[account_type].append(account)
    return deploy_targets

def backend_config_block_exists(context, path_dir):
    """
    Helper function to determine if a backend configuration block to enable an s3
    backend exists in the terraform directory

    This will return a boolean value
    """
    command = 'grep -nr ' + "'^[[:space:]][[:space:]]backend[[:space:]]\"s3\"[[:space:]]{}'" + ' {}/'.format(path_dir)
    process = context.run(command, warn=True)

    return bool(process.exited == 0)

def gen_backend_config_block(path_dir):
    """
    Helper function to create a terraform_backend.tf file and generate the
    terraform backend configuration block

    Params:
      tf_dir:  the terraform directory to init

    This will create a terraform_backend.tf file in the tf_dir location
    """
    LOGGER.info("terraform backend configuration block does not exist! Generating now!")
    file = open("{}/terraform_backend.tf".format(path_dir), "w")
    configuration = ['terraform {\n', '  backend \"s3\" {}\n', '}\n']
    file.writelines(configuration)
    file.close()

def terraform_generate_and_init(context, environment, target_config, backend_filepath, tf_dir):
    """
    Helper function to abstract the terraform backend config and terraform init
    portion of the process. This is useful as many phases need to do this step.

    Params:
      context:  an invoke context
      target_config:  an aws account we're targeting (summoner config)
      backend_filepath:  the filepath to write the (s3) backend config at
      tf_dir:  the terraform directory to init

    Raise:
      TerraformInitError - something in the init went wrong
    """
    def delayed_cleanup():
        """ tiny helper to ensure we always delete our backend file """
        if os.path.isfile(backend_filepath):
            os.remove(backend_filepath)

    # some simplifying local variables
    account = target_config['account_number']
    region = target_config['deploy_region']
    account_type = target_config['type']
    #check if terraform backend configuration block exists
    #if not...generate a file that contains the necessary empty configuration block
    if not backend_config_block_exists(context, tf_dir):
        gen_backend_config_block(tf_dir)
    # create the aws provider and
    # remote state configuration files
    gen_terraform_backend_config(target_config, backend_filepath, environment)

    # Next, we run a terraform init that grabs the providers locally and modules.

    fmt_result = context.run(f"terraform init -reconfigure -backend-config={backend_filepath} -input=false {tf_dir}", env=generate_tf_env_vars(target_config, environment), warn=True, hide=False)
    if fmt_result.exited != 0:
        LOGGER.error("Terraform initialize to Type:%s Account:%s Region:%s - Deployment FAILED.",
                     account_type, account, region)
        delayed_cleanup()
        raise TerraformInitFailure("Error initializing terraform (Type: {}, Acct:{}, Region:{})".format(
            account_type, account, region))

    delayed_cleanup()

def gitlab_push_tag(ci_project_id,
                    gl_private_token,
                    gl_url,
                    tag_value,
                    commit_sha):
    """
    This helper function makes a connection to the gitlab API and creates the
    desired version tag for a specified project

    Params:
        :param ci_project_id:       Gitlab project ID
        :param gl_private_token:    Gitlab private token
        :param gl_url:              Gitlab URL
        :param tag_value:           Tag name
        :param commit_sha:          Commit to tag
    """
    gitlab_setup = gitlab.Gitlab(gl_url, gl_private_token, api_version=4)
    project = gitlab_setup.projects.get(ci_project_id)
    project.tags.create({'tag_name': tag_value, 'ref': commit_sha})

def get_env_variables(config, environment):
    """
    This helper function sets the necessary environment variables for
    gitlab api calls

    Params:
        :param config:              Config file
        :param environment:         Environment variable dict
    """
    ci_project_id = None
    ci_commit_sha = None
    # Get the gitlab private token
    gl_token_env_var = config['summoner']['env_gitlab_api_token']
    try:
        gl_private_token = environment[gl_token_env_var]
    except:
        raise ConfigError("Unable to load GitLab private token from ENV variable {gl_token_env_var}")


    if environment.get("GITLAB_CI", None):

        # Grab and set environment Gitlab variables
        try:
            ci_project_id = environment['CI_PROJECT_ID']
        except:
            raise KeyError("Issue grabbing CI_PROJECT_ID environment variable")

        try:
            ci_commit_sha = environment['CI_COMMIT_SHA']
        except:
            raise KeyError("Issue grabbing commit SHA variable")


    return gl_private_token, ci_project_id, ci_commit_sha

def generate_tf_env_vars(target_config, environment):
    """
    This helper function sets the necessary environment variables for
    all terraform calls

    Params:
        :target_config:             config for target account
    """
    env_var_aws_key = environment.get(target_config['env_var_aws_key'])
    env_var_aws_secret = environment.get(target_config['env_var_aws_secret'])
    region = target_config['deploy_region']
    envir = {'AWS_ACCESS_KEY_ID': env_var_aws_key,
             'AWS_SECRET_ACCESS_KEY': env_var_aws_secret,
             'AWS_DEFAULT_REGION': region}
    return envir
