"""
Processes configuration for summoner
"""

import toml
from .common import Logger
# pylint: disable=W0107
# This is a very V1 approach here, but for now we will
# fill in default values manually and trust that the required
# options are present. After all, if the required ones are not
# present, something will fail and the user will notice.
#
# I'd really like to see us do some sort of config validator
# but that is out of scope for the V1 implementation of this.
LOGGER = Logger().get_logger(__name__)
DEFAULT_CONFIG_SETS = {
    '[[aws-account.terraform]]': {'state': {'dynamodb_table': 'ifrit_summoner_tfstate_locks',
                                            's3_bucket': 'ifrit-summoner-test-statefiles',
                                            's3_path': '/exampleproject/test/tf.state',
                                            's3_region': 'us-east-1'}},
}

def load_config(config_file="summoner.toml", o_validate_configs=True):
    """ Loads a summoner configuration file.

    This function will read process configs provided by the
    file (or file descriptor) provided. It will then
    validate the configs and return a ready-to-use configuration
    dictionary.

    While the auto validation can be disabled, that's really only
    intended to be used by test cases.

    Args:
        config_file (str, fd): the file descriptor/path to read from
        o_auto_validation (bool): turns on/off the config validation

    Returns:
        dict: a dictionary containing the configs

    """
    # try to load the config from file and parse into
    # a dictionary
    config = toml.load(config_file)
    LOGGER.debug("Config File: %s", config)

    # allow us to bypass config validation
    # really only useful for testing purposes
    if not o_validate_configs:
        return config

    validate_config(config)
    return process_config(config)


def process_config(config):
    """ Takes a loaded config dictionary and adds default values"

    This function will read process configs provided and
    fill in (if needed) any values not provided in the file with the
    default values above. It is intended to be used after a config
    file is loaded from disk.

    Args:
        config (dict): a dictionary of summoner configs

    Returns:
        dict: a dictionary containing the validated configs

    """

    # confirm the config has
    # the default values that we can provide.
    # this is awfully ugly.. but again, until we can put
    # a true config validator in, it's all we got.
    LOGGER.info("Starting to process config.")
    for account_data in config['aws-account']:
        if 'terraform' not in account_data:
            account_data['terraform'] = DEFAULT_CONFIG_SETS['[[aws-account.terraform]]'].copy()
            LOGGER.info("Using Default config set for aws account")

        if 'state' not in account_data['terraform']:
            account_data['terraform']['state'] = DEFAULT_CONFIG_SETS['[[aws-account.terraform]]']['state'].copy()
            print("added state")

        terraform_state = account_data['terraform']['state']

        if 'dynamodb_table' not in terraform_state:
            terraform_state['dynamodb_table'] = DEFAULT_CONFIG_SETS['[[aws-account.terraform]]']['state']['dynamodb_table']
        if 's3_bucket' not in terraform_state:
            terraform_state['s3_bucket'] = DEFAULT_CONFIG_SETS['[[aws-account.terraform]]']['state']['s3_bucket']
        if 's3_path' not in terraform_state:
            terraform_state['s3_path'] = DEFAULT_CONFIG_SETS['[[aws-account.terraform]]']['state']['s3_path']
        if 's3_region' not in terraform_state:
            terraform_state['s3_region'] = DEFAULT_CONFIG_SETS['[[aws-account.terraform]]']['state']['s3_region']
    LOGGER.info("Processing config complete.")
    # return the loaded config
    LOGGER.debug("%s", config)
    return config

#pylint: disable=unused-argument
# disabling the unused argument check until we fill this function out
def validate_config(config):
    """
    This (for now) is just a placeholder function for validating configs
    ideally, this function is used after any configuration is loaded, but
    it will be particularly useful for validating user-provided files as well
    as the configs we generate in test cases to ensure our tests do not diverge from
    our config requirements.

    But.. until we can do real validation, we'll just do a pass for now. Eventually
    we'll raise a ConfigError exception to inform the caller the config is invalid.
    """
    pass
