"""Summoner"""
import os
import time
#import toml

from .common import (
    ErrorCodes,
    TerraformInitFailure,
)
from .common import (
    calculate_deploy_targets,
    gen_terraform_backend_config,
    terraform_generate_and_init,
    generate_tf_env_vars
)
from .configs import load_config
from .common_logger import Logger

LOGGER = Logger().get_logger(__name__)

def run(context,
        config_file="summoner.toml",
        environment=os.environ,
        tf_dir="terraform",
        tf_automation=True,
        backend_filepath='_s3_backend.summoner',
        o_init_tf=True,
        o_apply_tf=True,
        account_type="test",
        o_destroy_tf=True):
    # pylint: disable=dangerous-default-value
    """
    This function will initialize terrafrom, run the apply to provison the infrastructure, perform integration tests and
    then destroy the infrastructure. If so desired it will leave the infrastructure up and running for a day.
    :param context:
    :param environment:
    :param tf_dir:
    :param tf_automation:
    :param o_init_tf:
    :param o_apply_tf:
    :param o_destroy_tf:
    :param o_integration_tests:
    :return:
    """
    # Leveraging TF_IN_AUTOMATION based on terraform docs to reduce noise:
    # https://www.terraform.io/guides/running-terraform-in-automation.html

    config = load_config(config_file)
    deploy_targets = calculate_deploy_targets(config)

    if tf_automation:
        environment['TF_IN_AUTOMATION'] = "On"

    for target_config in deploy_targets[account_type]:
    # Next, we run a terraform init that grabs the providers locally and modules.
        if o_init_tf:
            # attempt to run terraform init.
            # if an error occurs, re-classify it as a deployfailure (as it is in this context)
            try:
                terraform_generate_and_init(context, environment, target_config, backend_filepath, tf_dir)
            except TerraformInitFailure:
                return ErrorCodes.ERROR_TERRAFORM_INIT

        # This runs the actual terraform apply command to provision the infrastructure
        if o_apply_tf:
            override_tfvar_files = target_config.get('tfvar_override_files', list())

            # create the command line arguments for
            # terraform to override tfvar files. the expected
            # output here is:  "-var-file=fileone -var-file=filetwo ..etc"
            override_command = "".join([" -var-file={}".format(tfvar_file) for tfvar_file in override_tfvar_files])

            start = time.time()
            gen_terraform_backend_config(target_config, backend_filepath, environment)
            fmt_result = context.run(f"terraform apply -input=false -no-color -auto-approve {override_command} {tf_dir}", env=generate_tf_env_vars(target_config, environment), warn=True, hide=False)
            if os.path.isfile(backend_filepath):
                os.remove(backend_filepath)
            if fmt_result.exited != 0:
                end = time.time()
                LOGGER.info("Terraform failed apply in: " + "%0.2f" % (end - start) + " seconds.")
                return ErrorCodes.ERROR_TERRAFORM_APPLY
            # removing the backend config file from disk
            end = time.time()
            LOGGER.info("Terraform Applied in: " + "%0.2f" % (end - start) + " seconds.")

        # Integration testing will be implemented here
        # arguments: o_persist_infra_duration=43200 & o_integration_tests=True
        # if o_integration_tests:
        #     time.sleep(o_persist_infra_duration)
        if o_destroy_tf:
            override_tfvar_files = target_config.get('tfvar_override_files', list())

            # create the command line arguments for
            # terraform to override tfvar files. the expected
            # output here is:  "-var-file=fileone -var-file=filetwo ..etc"
            override_command = "".join([" -var-file={}".format(tfvar_file) for tfvar_file in override_tfvar_files])
            environment['TF_WARN_OUTPUT_ERRORS'] = "1"
            start = time.time()
            gen_terraform_backend_config(target_config, backend_filepath, environment)
            fmt_result = context.run(f"terraform destroy -input=false -auto-approve {override_command} {tf_dir}", env=generate_tf_env_vars(target_config, environment), warn=True, hide=False)
            # removing the backend config file from disk
            if os.path.isfile(backend_filepath):
                os.remove(backend_filepath)
            end = time.time()
            LOGGER.info("Terraform Destroyed in: " + "%0.2f" % (end - start) + " seconds.")
            if fmt_result.exited != 0:
                LOGGER.error("Terraform Destroy pass failed. \n %s", fmt_result.stdout)
                return ErrorCodes.ERROR_TERRAFORM_DESTROY

    return ErrorCodes.SUCCESS
