"""Summoner"""
import os
import copy
import re

import gitlab

from .common import (
    ConfigError,
    DeployFailure,
    ErrorCodes,
    NotificationType,
    RollbackComplete,
    RollbackFailure,
    TerraformInitFailure,
)
from .common import (
    calculate_deploy_targets,
    gen_terraform_backend_config,
    get_env_variables,
    gitlab_push_tag,
    send_pd_notification,
    terraform_generate_and_init,
    generate_tf_env_vars

)
from .configs import load_config
from .common_logger import Logger

LOGGER = Logger().get_logger(__name__)

# disabling the pylint too-many-returns here because
# this function is intended to have many return calls
# to exit with the correct error code. We could refactor this
# out into multiple helper functions, but that's for a later date.
#
# disabling pylint's too many branches check because this function
# has a lot of if-statements. While we could refactor this, it would
# only make it more complicated. Keeping the if-statements as is.
#
# disabling the too many local variables check because of the default
# character length on function names
#
#

def run(context, #pylint: disable=too-many-return-statements,too-many-branches,too-many-locals,too-many-statements
        config_file="summoner.toml",
        tf_dir="terraform",
        stage_account_type="stage",
        prod_account_type="prod",
        environment=copy.deepcopy(os.environ),
        deploy_tag_prefix="summoner-",
        backend_filepath='_s3_backend.summoner',
        gl_url="https://gitlab.bit9.local",
        o_launch_to_stage=True,
        o_launch_to_prod=False,
        o_rollback_disabled=False):
    """
    The validate stage function runs a terraform apply to the specified stage
    environment in all region(s) specified.

    Params:
      context:             an invoke context for running commands
      config_file:         a path or file descriptor we should read the config from
      tf_dir:              the directory where the terraform files reside
      stage_account_type:  the name of the type for stage accounts
      prod_account_type:   the name of the type for prod accounts
      environment:         a dictionary of environment variables we can reference
      deploy_tag_prefix:   the prefix for git tags we'll apply on successful launch
      backend_filepath:    the filepath to place the generated s3 backend config for terraform

      o_launch_to_stage:   if true, apply code to the specified stage accounts
      o_launch_to_prod:    if true, apply code to the specified prod accounts
      o_rollback_disabled:  if true, rollback functionality is disabled

    Returns 0 if successful, or a integer error code if failure.
    """
    # V2: add ability to ensure no other master-branch-based pipeline are executing

    # load up the configs
    LOGGER.info("Starting LAUNCH phase.")
    config = load_config(config_file)

    # Get the gitlab private token, project_id and commit sha
    try:
        gl_private_token, ci_project_id, ci_commit_sha = get_env_variables(config, environment)
    except ConfigError:
        return ErrorCodes.ERROR_MISSING_CONFIG
    except KeyError:
        return ErrorCodes.ERROR_MISSING_PARAMS

    # get pagerduty keys:
    try:
        pd_low_key = environment[config['summoner']['env_pagerduty_key_low']]
        pd_high_key = environment[config['summoner']['env_pagerduty_key_high']]

        # confirm pagerduty keys are accessible
        if len(pd_low_key) < 5 or len(pd_high_key) < 5:
            raise ValueError("PagerDuty keys are too short, please double check the "
                             "keys are currently set as environment variables. Trying "
                             f"to read ENV variables for low:{pd_low_key} and "
                             f"high:{pd_high_key} alerts.")
    except KeyError as excinfo:
        raise KeyError("Unable to read PagerDuty keys from environment variables. Are they "
                       "specified in the summoner configuration file?") from excinfo
    # get all regions we need to deploy to, organized by
    # the type of account test/stage/prod
    LOGGER.info("Calculating Deploy Targets.")
    deploy_targets = calculate_deploy_targets(config)
    LOGGER.info("Calculating deploy target completed")
    # git tag prefixes
    stage_prefix = deploy_tag_prefix + stage_account_type + "-r"
    prod_prefix = deploy_tag_prefix + prod_account_type + "-r"
    LOGGER.info("Calculating last deploy versions.")
    # find/calculate the last deploy versions and next version id
    stage_version_id = find_last_deploy(context, stage_prefix)
    prod_version_id = find_last_deploy(context, prod_prefix)

    # increase our deployment version number by one more than
    # the greater of stage's or prod's larger number. This will
    # allow us to continue increasing the deployment number
    # even if prod isn't deployed.  If they are deployed together
    # though, the number will be the same.
    LOGGER.info("Increasing deployment version number by 1")
    next_version_id = 1 + max(
        stage_version_id if stage_version_id is not None else 0,
        prod_version_id if prod_version_id is not None else 0,
    )

    # Helper function (see inline comment below)
    def helper_run_deploy(context, account_type, deploy_targets,
                          git_tag_prefix, tf_dir, next_version_id, environment):
        """
        Helper function to reduce code duplication as we deploy
        to both stage and production accounts here.
        """
        try:
            LOGGER.info(f"Beginning deployment to {account_type.upper()} accounts")
            deploy(context,
                   account_type=account_type,
                   deploy_targets=deploy_targets,
                   git_tag_prefix=git_tag_prefix,
                   backend_filepath=backend_filepath,
                   tf_dir=tf_dir,
                   environment=environment,
                   o_rollback_disabled=o_rollback_disabled)

            # if we're here the deployment was a success, so
            # we'll create a git tag to celebrate
            tag_value = git_tag_prefix + str(next_version_id)
            tag_result = context.run(f"git tag {tag_value}", warn=True)
            if tag_result.exited != 0:
                raise Exception(f"Unable to tag {account_type} deploy version {tag_value} in git.")
            if not o_rollback_disabled:
                try:
                    gitlab_push_tag(ci_project_id,
                                    gl_private_token,
                                    gl_url,
                                    tag_value,
                                    ci_commit_sha)
                except gitlab.GitlabAuthenticationError as error:
                    LOGGER.error("Authentication to GitLab server at %s failed: %s", gl_url, error)
                    return ErrorCodes.ERROR_GITLAB

            return ErrorCodes.SUCCESS

        except RollbackComplete:
            LOGGER.exception(f"A deploy to an account of type:{account_type} failed, "
                             "but rollback completed successfully. Failing the pipeline.")
            return ErrorCodes.ERROR_LAUNCH_RECOVERY

        except RollbackFailure:
            LOGGER.exception(f"FAILURE in rollback of {account_type.upper()} account(s). Escalating to a human.")
            if account_type == 'prod':
                send_pd_notification(pd_high_key,
                                     NotificationType.HIGH_PRIORITY,
                                     environment["CI_PROJECT_PATH_SLUG"],
                                     f"FAILURE in rollback of {account_type.upper()} account(s)")
            else:
                send_pd_notification(pd_low_key,
                                     NotificationType.LOW_PRIORITY,
                                     environment["CI_PROJECT_PATH_SLUG"],
                                     f"FAILURE in rollback of {account_type.upper()} account(s)")

            return ErrorCodes.ERROR_LAUNCH_FAILURE

        except DeployFailure:
            LOGGER.exception(f"A deploy to an account of type:{account_type} failed, "
                             "Rollback was disabled. Failing the pipeline and escalating to a human.")
            if account_type == 'prod':
                send_pd_notification(pd_high_key,
                                     NotificationType.HIGH_PRIORITY,
                                     environment["CI_PROJECT_PATH_SLUG"],
                                     f"FAILURE in deploy of {account_type.upper()} account(s)")
            else:
                send_pd_notification(pd_low_key,
                                     NotificationType.LOW_PRIORITY,
                                     environment["CI_PROJECT_PATH_SLUG"],
                                     f"FAILURE in deploy of {account_type.upper()} account(s)")
            return ErrorCodes.ERROR_TERRAFORM_APPLY

    LOGGER.info("Deploying to stage.")
    # DEPLOY TO STAGE
    if o_launch_to_stage:
        return_value = helper_run_deploy(
            context=context,
            account_type=stage_account_type,
            deploy_targets=deploy_targets,
            git_tag_prefix=stage_prefix,
            tf_dir=tf_dir,
            next_version_id=next_version_id,
            environment=environment)
        if return_value != ErrorCodes.SUCCESS:
            return return_value

    LOGGER.info("Deploying to prod.")
    # DEPLOY TO PROD (only if stage was successful)
    if o_launch_to_prod:
        return_value = helper_run_deploy(
            context=context,
            account_type=prod_account_type,
            deploy_targets=deploy_targets,
            git_tag_prefix=prod_prefix,
            tf_dir=tf_dir,
            next_version_id=next_version_id,
            environment=environment)
        if return_value != ErrorCodes.SUCCESS:
            return return_value

    # If we got here, all is well with the world.
    return ErrorCodes.SUCCESS


def deploy(context, account_type, deploy_targets, tf_dir,
           backend_filepath, git_tag_prefix, environment, o_rollback_disabled):
    """
    For the provided account_type, attempt to deploy to all targets
    in deploy_targets[account_type].

    Params:
      context: an invoke context
      account_type:  e.g. "stage" or "prod" to target
      deploy_targets:  list of all available targets.
      tf_dir: the path to the root of the terraform code path
      backend_filepath: the filename for the s3 backend config we write to disk
      git_tag_prefix: prefix to use with our git tagging
      o_rollback_disabled: disables rollback, default False

    Raises:
      DeployFailure - when a deploy fails, and there is no rollback target
      RollbackComplete - when a deploy fails, but the rollback succeeds
      RollbackFailure - when a deploy AND the rollback both fail (BAD)
    """

    last_deploy_id = find_last_deploy(context, git_tag_prefix)
    rollback_git_ref = git_tag_prefix + str(last_deploy_id)

    # Attempt to do the deployment to the Stage accounts.
    # Wrapping this in a try-except block so that we can
    # correctly begin a rollback if something fails.
    try:

        # begin stage deployment loop. We're going to deploy to each
        # stage environment specified in the config
        attempted_deploy_targets = list()
        target = None

        for target in deploy_targets[account_type]:

            # track which targets we attempt to deploy to, makes rollback much easier
            attempted_deploy_targets.append(target)
            deploy_to_target(context, target, backend_filepath, tf_dir, None, environment)
            LOGGER.info("Successfully deployed to type:%s, account:%s, region:%s",
                        account_type, target['account_number'], target['deploy_region'])


    # Intentionally using a bare exception here because
    # we MUST attempt rollback on any possible error.
    except:  # pylint: disable=bare-except

        # Rollback has been disabled
        if o_rollback_disabled:
            failed_target = target  # for error reporting
            LOGGER.info("Rollback has been disabled. Halting attempts at rollback.")
            raise DeployFailure("Deploy failed and no rollback.")

        failed_target = target  # for error reporting
        LOGGER.exception("FAILURE in deploying: %s %s %s.  Attempting rollback.",
                         account_type, failed_target['account_number'], failed_target['deploy_region'])

        # something went wrong! Attempt to rollback.
        # but first, check we actually have something to rollback to..
        if last_deploy_id is None:
            LOGGER.warning("No rollback ID found. This is a new project? Halting attempts at rollback.")
            raise DeployFailure("Deploy failed and no rollback targets.")


        # attempt to roll the environment back
        try:

            # we'll keep track of the targets we successfully
            # rolled back. Useful to know which (if any) env's broke
            targets_rolled_back = list()

            # attempt to rollback the targets in the reverse
            # order that we deployed them (so we'll try the failed target first)
            for target in attempted_deploy_targets[::-1]:
                deploy_to_target(context, target, backend_filepath, tf_dir, rollback_git_ref, environment)
                targets_rolled_back.append(target)

        # Intentionally using a bare exception here because
        # we should return a RollbackFailure exception so upstream
        # users know something bad happened here.
        except Exception as exec_info:
            raise RollbackFailure(f"BAD: Unable to rollback. Error deploying {account_type},"
                                  f"{target['account_number']},{target['deploy_region']}: {exec_info}")

        # If rollback was successful, still raise an exception as the original
        # deploy failed
        raise RollbackComplete(f"Deploy failed for {account_type}:{failed_target['account_number']}. "
                               f"Rollback to {rollback_git_ref} successful.")

    # this comment here just to help indicate the close of this python function

def find_last_deploy(context, deploy_tag_prefix):
    """
    A helper function to review the git tags present in the repository and
    find the most-recent deployed summoner code. That git commit will be
    the one we re-apply (rollback to) if the current deploy attempt fails. If
    the deploy succeeds, then we can increment the deploy version and tag the
    new commit appropriately.

    Params:
      context: an invoke context that can be used to run shell commands
      deploy_tag_prefix:  prefix of the git tag to indicate it's a summoner deploy indicator

    Returns:
      (None) if there is no previous commit (no rollback possible)
      (Int) last deploy number

    Raises:
      RuntimeError - if the git command had a failure
    """

    # Get all git tags in the current repo..
    version_tags = list()
    LOGGER.debug("Attempting to get git tags in the current repo.")
    git_ref = context.run("git show-ref --tags", hide=True, warn=True)

    if git_ref.exited != 0:
        return None
    LOGGER.info("Attempting to iterate through tags")
    # Look for all tags which match our version tag syntax (prefix + version_int)
    for line in git_ref.stdout.split('\n'):
        match_result = re.search('refs/tags/' + deploy_tag_prefix + '([0-9]+)$', line)
        if match_result is not None:
            version_tags.append(int(match_result.group(1)))
    LOGGER.info("Tags matched, attempting to sort")
    version_tags = sorted(version_tags)

    # If we didn't find any tags, then return None
    # if we found tags, grab the highest integer
    # and return it.
    if not version_tags:
        return None
    LOGGER.info("Returning highest integer version")
    return version_tags[-1]



def deploy_to_target(context, target_config, backend_filepath, tf_dir, gitref, environment):
    """
    Invoke terraform to deploy to the targeted accunts and regions.

    This function will invoke terraform on using the provided target configuration. We will write an appropriate
    tf file (json format) for the terraform remote state and aws provider portion. Using the config data this
    function will then invoke terraform with the correct ordering of tffiles.

    NOTE: this function will NOT handle any kind of rollback. At it's most basic level, this function is
    intended as a terraform wrapper.

    Args:
      context: an invoke context for running shell commands
      target_config: config for an aws account to deploy to. One config per (type, account, region) combination.
      backend_filepath:  the filename for the backend filepath we write to disk
      tf_dir: the directory containing the root of the terraform to execute
      gitref: the git reference to the terraform code to deploy. If None, uses the current git checkout.

    Raises:
      ValueError: if git is unable to checkout the branch/tag/commit provided
      DeployFailure: if an error in deploy occurs

    """
    # we'll need to do this for each region in the
    # specified stage environment
    account = target_config['account_number']
    region = target_config['deploy_region']
    account_type = target_config['type']
    override_tfvar_files = target_config.get('tfvar_override_files', list())

    LOGGER.info("Account:%s Region:%s - Starting deployment using git ref: %s.", account, region, gitref)

    # checkout whichever git reference we need to.
    # if the gitref is None, we'll just use the repo as-is
    if gitref is not None:
        git_result = context.run("git checkout {}".format(gitref), warn=True)
        if git_result.exited != 0:
            raise ValueError("Unable to checkout git reference: {}".format(gitref))

    # create the command line arguments for
    # terraform to override tfvar files. the expected
    # output here is:  "-var-file=fileone -var-file=filetwo ..etc"
    override_command = "".join([" -var-file={}".format(tfvar_file) for tfvar_file in override_tfvar_files])

    # attempt to run terraform init.
    # if an error occurs, re-classify it as a deployfailure (as it is in this context)
    try:
        terraform_generate_and_init(context, environment, target_config, backend_filepath, tf_dir)
    except TerraformInitFailure as excinfo:
        raise DeployFailure(str(excinfo))

    # create the aws provider and
    # remote state configuration files
    gen_terraform_backend_config(target_config, backend_filepath, environment)

    # to begin, we run a terraform apply
    fmt_result = context.run(f"terraform apply -no-color -auto-approve -input=false {override_command} {tf_dir}", env=generate_tf_env_vars(target_config, environment), warn=True, hide=False)

    # removing the backend config file from disk
    if os.path.isfile(backend_filepath):
        os.remove(backend_filepath)

    # if the deploy failed, raise an exception. This will allow other
    # code to notify and/or do a rollback/re-deploy if desired.
    if fmt_result.exited != 0:
        LOGGER.error("Terraform apply to Type:%s Account:%s Region:%s - Deployment FAILED.",
                     account_type, account, region)
        raise DeployFailure("Error applying terraform (Type: {}, Acct:{}, Region:{})".format(
            account_type, account, region))
