"""Summoner"""
import copy
import os
import gitlab

from .common import (
    ErrorCodes,
    TerraformInitFailure,
)
# YO! remove me when pylint releases import fix
from .common import (
    calculate_deploy_targets,
    gen_terraform_backend_config,
    terraform_generate_and_init,
    generate_tf_env_vars
)
from .configs import load_config
from .common_logger import Logger



LOGGER = Logger().get_logger(__name__)

# disabling the pylint too-many-returns here because
# this function is intended to have many return calls
# to exit with the correct error code. We could refactor this
# out into multiple helper functions, but that's for a later date.
#
# disabling pylint's too many branches check because this function
# has a lot of if-statements. While we could refactor this, it would
# only make it more complicated. Keeping the if-statements as is.
#
# and now disabling the too many local variables check because pylint
# seems to hate long functions with human-readble variables. No.. I'm not bitter.
#
def run(context,   #pylint: disable=too-many-return-statements,too-many-branches,too-many-locals
        config_file="summoner.toml",
        environment=copy.deepcopy(os.environ),
        gl_url="https://gitlab.bit9.local",
        tf_dir="terraform",
        tf_automation=True,
        backend_filepath='_s3_backend.summoner',
        plan_account_type="stage",
        o_init_tf=True):
    """
     The plan task runs a plan of the terraform code against an environment.
     It also supports logic to determine if you are running summoner locally or in a gitlab
     pipeline and prints the output to an MR or st.out accordingly

     Params:
        :param context:            Invoke context
        :param config_file:        The path to the configuration file
        :param environment:        Environment varibles dictionary
        :param gl_url:             URL of the Gitlab server
        :param tf_dir:             Terraform directory name
        :param tf_automation:      Enable Terraform Automation
        :param plan_account_type:  The account type to process from the config file ("stage" "prod" etc)
        :param o_init_tf:          Enable Terraform initialization

     Returns 0 if successful, or a integer error code if failure.
    """
    # load up the configs
    LOGGER.info("Starting PLAN phase.")
    config = load_config(config_file)

    # Get the gitlab private token
    gl_token_env_var = config['summoner']['env_gitlab_api_token']
    try:
        gl_private_token = environment[gl_token_env_var]
    except KeyError:
        LOGGER.error("Unable to load GitLab private token from ENV variable {gl_token_env_var}")
        return ErrorCodes.ERROR_MISSING_CONFIG

    if environment.get("GITLAB_CI", None):

        # Grab and set environment Gitlab variables
        try:
            ci_project_id = environment['CI_PROJECT_ID']
        except KeyError:
            LOGGER.error("Issue grabbing CI_PROJECT_ID environment variable")
            return ErrorCodes.ERROR_MISSING_PARAMS

        try:
            ci_commit_sha = environment['CI_COMMIT_SHA']
        except KeyError:
            LOGGER.error("Issue grabbing commit SHA variable")
            return ErrorCodes.ERROR_MISSING_PARAMS

    # Leveraging TF_IN_AUTOMATION based on terraform docs to reduce noise:
    # https://www.terraform.io/guides/running-terraform-in-automation.html

    if tf_automation:
        environment['TF_IN_AUTOMATION'] = "On"
    plan_results = ""

    # for each stage environment, do a plan to confirm
    # that only the expected changes will take place.
    deploy_targets = calculate_deploy_targets(config)
    for target_config in deploy_targets[plan_account_type]:

        # some simplifying local variables
        account = target_config['account_number']
        region = target_config['deploy_region']
        account_type = target_config['type']

        # attempt to run terraform init.
        if o_init_tf:
            try:
                terraform_generate_and_init(context, environment, target_config, backend_filepath, tf_dir)
            except TerraformInitFailure as excinfo:
                LOGGER.error("Terraform Initialize pass failed. \n %s", str(excinfo))
                return ErrorCodes.ERROR_TERRAFORM_INIT

        # generate the backend_filepath for our plan
        gen_terraform_backend_config(target_config, backend_filepath, environment)

        # generate the override command lines
        override_tfvar_files = target_config.get('tfvar_override_files', list())
        override_command = " ".join(["-var-file={}".format(tfvar_file) for tfvar_file in override_tfvar_files])
        LOGGER.info("'" + override_command + "'")

        # run the terraform plan
        fmt_result = context.run(f"terraform plan -no-color {override_command} {tf_dir}", env=generate_tf_env_vars(target_config, environment), warn=True, hide=False)

        if os.path.isfile(backend_filepath):  # cleanup the backend file is something failed
            os.remove(backend_filepath)

        # see if the plan was successful
        if fmt_result.exited != 0:
            LOGGER.error("Terraform plan failed failed. Terraform error: %s", fmt_result.stderr)
            return ErrorCodes.ERROR_TERRAFORM_PLAN

        # Get the commit and make comment
        header_line = f"Created Terraform Plan for Type:{account_type}, Acct:{account}, Region:{region}"
        plan_results = header_line + " :\n```python\n"+fmt_result.stdout+"```"
        if environment.get("GITLAB_CI", None):
            # Initialize Gitlab information
            try:
                gitlab_comment_on_commit(ci_project_id=ci_project_id,
                                         ci_commit_sha=ci_commit_sha,
                                         gl_url=gl_url,
                                         gl_private_token=gl_private_token,
                                         plan_results=plan_results)
            except gitlab.GitlabAuthenticationError as error:
                LOGGER.error("Something went wrong with Gitlab %s", error)
                return ErrorCodes.ERROR_GITLAB
        else:
            LOGGER.info(plan_results)

    # V2 add in tag/policy enforcement
    return ErrorCodes.SUCCESS

def gitlab_comment_on_commit(ci_commit_sha,
                             ci_project_id,
                             gl_private_token,
                             gl_url,
                             plan_results):
    """
    This helper function makes a connection to gitlab API and posts a comment with the output
     of the Terraform plan to an MR.

    Params:
        :param ci_commit_sha:       Gitlab commit sha
        :param ci_project_id:       Gitlab project ID
        :param gl_private_token:    Gitlab private token
        :param gl_url:              Gitlab URL
        :param plan_results:        Results from Terraform plan
    """
    gitlab_setup = gitlab.Gitlab(gl_url, gl_private_token, api_version=4)
    project = gitlab_setup.projects.get(ci_project_id)
    commit = project.commits.get(ci_commit_sha)
    commit.comments.create({'note': plan_results})
