"""Summoner"""
import os
import copy

from .common import (
    ErrorCodes,
    TerraformInitFailure,
)
from .common import (
    calculate_deploy_targets,
    terraform_generate_and_init,
)
from .configs import load_config
from .common_logger import Logger

LOGGER = Logger().get_logger(__name__)

def run(context,
        branch_name,
        account_type='test',
        backend_filepath='_s3_backend.summoner',
        config_file="summoner.toml",
        master_branch_name="master",
        tf_dir="terraform",
        o_check_format=True,
        o_check_branch_diverge=True,
        o_init_tf=True,
        o_run_tflint=True,
        o_validate_tf=True,
        environment=copy.deepcopy(os.environ)):
    """
    The validate run function runs a series of static analysis tools on the
    provided terraform code.

    Params:
      branch_name:  the name of the git branch we're on
      master_branch_name:  the name of the "master" branch to compare divergence against
      tf_dir:  the directory where the terraform files reside

      o_check_format:          enable terraform fmt check
      o_check_branch_diverge:  enable branch divergance check
      o_init_tf:               enable terraform init
      o_validate_tf:           enable terraform validate check
      o_run_tflint:            enable tflint pass

    Returns 0 if successful, or a integer error code if failure.
    """
    # load up the configs
    LOGGER.info("Starting Validate phase.")
    config = load_config(config_file)
    LOGGER.info("Calculating Deploy Targets.")
    deploy_targets = calculate_deploy_targets(config)
    LOGGER.debug("Deploy targets calculated: %s", deploy_targets)

    # First up, we check the formatting on the terraform file(s).
    if o_check_format:
        fmt_result = context.run("terraform fmt -check=true", warn=True, hide=True)
        if fmt_result.exited != 0:
            LOGGER.error("Terraform Format pass failed. "
                         "Files are not correctly formatted: %s", fmt_result.stdout)
            return ErrorCodes.ERROR_TERRAFORM_FORMAT

    # Follow that up with a tflint execution.
    if o_run_tflint:
        fmt_result = context.run("tflint", warn=True, hide=True)
        if fmt_result.exited != 0:
            LOGGER.error("tflint pass failed. \n %s", fmt_result.stdout)
            return ErrorCodes.ERROR_TF_LINT

    # Next, let's confirm the branch is still in sync with the master branch
    # don't want to be running terraform code if the master branch has something
    # newer.
    if o_check_branch_diverge:
        fmt_result = context.run(f"git rev-list remotes/origin/{branch_name}..remotes/origin/{master_branch_name} --", warn=True, hide=True)
        if fmt_result.stdout:
            LOGGER.error("This branch is missing commits from %s. Please merge that branch into this one."
                         "Commits missing: \n %s", master_branch_name, fmt_result.stdout)
            return ErrorCodes.ERROR_BRANCH_DIVERGED

    for target_config in deploy_targets[account_type]:
    # Do the init (if it's turned on by the option flag)!
        if o_init_tf:
            # attempt to run terraform init.
            # if an error occurs, re-classify it as a deployfailure (as it is in this context)
            try:
                terraform_generate_and_init(context, environment, target_config, backend_filepath, tf_dir)
            except TerraformInitFailure:
                LOGGER.exception(f"FAILURE in Terraform Initialization of {account_type.upper()} account(s).")
                return ErrorCodes.ERROR_TERRAFORM_INIT

        # Next, we run a terraform validate.
        if o_validate_tf:
            override_tfvar_files = target_config.get('tfvar_override_files', list())

            # create the command line arguments for
            # terraform to override tfvar files. the expected
            # output here is:  "-var-file=fileone -var-file=filetwo ..etc"
            override_command = "".join([" -var-file={}".format(tfvar_file) for tfvar_file in override_tfvar_files])
            fmt_result = context.run(f"terraform validate {override_command} {tf_dir}", warn=True, hide=False)
            if fmt_result.exited != 0:
                LOGGER.error("Terraform Validate pass failed.\n")
                return ErrorCodes.ERROR_TERRAFORM_VALIDATE

    # V2 add in tag/policy enforcement
    return ErrorCodes.SUCCESS
