"""Summoner

This file acts as a bridge between invoke tasks and the various
functions we want to execute with it. The goal here is to keep
this file as clean as possible and it simply dispatches requests
out to seperate files.

"""
from invoke import task
from invoke.exceptions import Exit

from src.summoner.tasks_common import param_helper

import src.summoner.validate as m_validate
import src.summoner.plan as m_plan
import src.summoner.launch as m_launch
import src.summoner.dryrun as m_dryrun

@task
def validate(context, args="", kwargs=""):
    """
    Validate the terraform is valid using static analysis checks.
    See the validate.run function for more detail.
    NOTE: need to put the 'context' parameter explictly because invoke gets mad otherwise.

    Params:
      args - a comma separated list of arguments to pass into sub-function
      kwargs - a comma separted list of key=value pairs to pass to sub-function
    """
    params = param_helper(args, kwargs)
    return_code = m_validate.run(context, *params[0], **params[1])
    print(return_code)
    raise Exit(code=return_code)

@task
def plan(context, args="", kwargs=""):
    """
    Generate a terraform plan and post it to the branch's MR for review.
    See the plan.run function for mroe detail.
    NOTE: need to put the 'context' parameter explictly because invoke gets mad otherwise.

    Params:
      args - a comma separated list of arguments to pass into sub-function
      kwargs - a comma separted list of key=value pairs to pass to sub-function
    """
    params = param_helper(args, kwargs)
    return_code = m_plan.run(context, *params[0], **params[1])
    print(return_code)
    raise Exit(code=return_code)

@task
def dryrun(context, args="", kwargs=""):
    """
    Generate a terraform apply to the sandbox account.
    See the dryrun.run function for more detail.
    NOTE: need to put the 'context' parameter explicitly because invoke gets mad otherwise.

    Params:
      args - a comma separated list of arguments to pass into sub-function
      kwargs - a comma separted list of key=value pairs to pass to sub-function
    """
    params = param_helper(args, kwargs)
    return_code = m_dryrun.run(context, *params[0], **params[1])
    print(return_code)
    raise Exit(code=return_code)

@task
def launch(context, args="", kwargs=""):
    """
    Act on a terraform plan and run an apply to the staging env in the target region(s).
    See the stage.run function for mroe detail.
    NOTE: need to put the 'context' parameter explictly because invoke gets mad otherwise.

    Params:
      args - a comma separated list of arguments to pass into sub-function
      kwargs - a comma separted list of key=value pairs to pass to sub-function
    """
    params = param_helper(args, kwargs)
    return_code = m_launch.run(context, *params[0], **params[1])
    print(return_code)
    raise Exit(code=return_code)
