"""
These test cases are designed to validate the proper functionality of
common functions
"""
import filecmp

from code.src.summoner.common import (
    DuplicateDeployCombination,
    NotificationType,
    PagerDutyNotificationError,
    ConfigError,
    send_pd_notification,
    set_severity_level,
    gen_terraform_backend_config,
    calculate_deploy_targets,
    get_env_variables,
    backend_config_block_exists,
)

import pytest
import mock
from invoke import MockContext, Result

@mock.patch('pypd.EventV2.create')
def test_send_pd_notification(mock_pypd_eventv2_create):
    """Testing notification Exception is not raised"""
    mock_pypd_eventv2_create.return_value = {
        "status": "success",
        "message": "Event processed",
        "dedup_key": "01234"
    }
    summary = "Testing low priority notification"
    source = "Summoner: Test"
    try:
        send_pd_notification("ff0750", NotificationType.LOW_PRIORITY, source, summary)
    except PagerDutyNotificationError as error:
        pytest.fail("send_pd_notification() raised exception '{}' unexpectedly!".format(error))

@mock.patch('pypd.EventV2.create')
def test_send_pd_notification_fail(mock_pypd_eventv2_create):
    """Testing notification Exception is raised"""
    mock_pypd_eventv2_create.return_value = {
        "status": "invalid event",
        "message": "Event object is invalid",
        "errors": [
            "Length of 'routing_key' is incorrect (should be 32 characters)"
        ]
    }
    summary = "Testing low priority notification"
    source = "Summoner: Test"
    with pytest.raises(PagerDutyNotificationError) as excinfo:
        send_pd_notification("ff0750", NotificationType.LOW_PRIORITY, source, summary)
    assert "PagerDuty notification event error. Recieved status:" in str(excinfo.value)

def test_set_severity_level_low():
    """Test setting PagerDuty notification severity level to low"""
    severity = set_severity_level(NotificationType.LOW_PRIORITY)
    assert severity == "info"

def test_set_severity_level_high():
    """Test setting PagerDuty notification severity level to high"""
    severity = set_severity_level(NotificationType.HIGH_PRIORITY)
    assert severity == "critical"

def test_calculate_deploy_targets():
    """Testing duplicate deploy target Exception is not raised"""
    config = {'aws-account': [
        {'type': 'Test',
         'account_number': '123',
         'deploy_region': 'us-east-1',
         'env_var_aws_key': 'abcd',
         'env_var_aws_secret': '1234'
         },
        {'type': 'Stage',
         'account_number': '4567',
         'deploy_region': 'us-east-1',
         'env_var_aws_key': 'efgh',
         'env_var_aws_secret': '5678'}
        ]}

    deploy_targets = calculate_deploy_targets(config)
    assert len(deploy_targets) == 2
    assert deploy_targets['Test'][0]['account_number'] == '123'
    assert deploy_targets['Stage'][0]['account_number'] == '4567'

def test_calculate_deploy_targets_environment_key():
    """Testing environment specifier for deploy target Exception is not raised"""
    config = {'aws-account': [
        {'type': 'Test', # Test WITH deploy_environment specifier
         'account_number': '123',
         'deploy_region': 'us-east-1',
         'deploy_environment': 'enviad',
         'env_var_aws_key': 'abcd',
         'env_var_aws_secret': '1234'
         },
        {'type': 'Test', # Test WITH deploy_environment specifier same account and different deploy environment
         'account_number': '123',
         'deploy_region': 'us-east-1',
         'deploy_environment': 'enviad01',
         'env_var_aws_key': 'abcd',
         'env_var_aws_secret': '1234'
        },
        {'type': 'Stage', # Test WITHOUT deploy_environment specifier
         'account_number': '4567',
         'deploy_region': 'us-east-1',
         'env_var_aws_key': 'efgh',
         'env_var_aws_secret': '5678'}
        ]}

    deploy_targets = calculate_deploy_targets(config)
    assert len(deploy_targets) == 2
    assert 'deploy_environment' in deploy_targets['Test'][0].keys()
    assert deploy_targets['Test'][0]['deploy_environment'] == 'enviad'
    assert deploy_targets['Test'][1]['deploy_environment'] == 'enviad01'
    assert 'deploy_environment' not in deploy_targets['Stage'][0].keys()

def test_calculate_deploy_targets_duplicate():
    """Testing duplicate deploy target Exception is raised"""
    config = {'aws-account': [
        {'type': 'Test',
         'account_number': '0123',
         'deploy_region': 'us-east-1',
         'env_var_aws_key': 'abcd',
         'env_var_aws_secret': '1234'},
        {'type': 'Test',
         'account_number': '0123',
         'deploy_region': 'us-east-1',
         'env_var_aws_key' : 'abcd',
         'env_var_aws_secret': '1234'}
        ]}
    with pytest.raises(DuplicateDeployCombination) as excinfo:
        calculate_deploy_targets(config)
    assert "Received a duplicate" in str(excinfo.value)

def test_gen_terraform_backend_config(tmpdir):
    """Testing the generation of a terraform backend config file"""
    config = {'terraform': {'state':{'s3_bucket': 'tf_statefiles',
                                     's3_region': 'us-east-1',
                                     's3_path': 'cb/stage/tf.state',
                                     'dynamodb_table': 'tfstate_locks'}}}
    envir = {'AWS_ACCESS_KEY_ID': 'ASASGETXXCD',
             'AWS_SECRET_ACCESS_KEY': 'asdfhgbe23d',
             'AWS_DEFAULT_REGION': 'us-east-1'}
    filename = tmpdir.join('test_summoner_terraform.tf')
    gen_terraform_backend_config(config, filename, envir)

    #Create a config file manually
    compare_file = tmpdir.join('compare.tf')
    out = open(compare_file, "w")
    out.writelines(['region = "us-east-1"\n', 'bucket = "tf_statefiles"\n',
                    'key = "cb/stage/tf.state"\n', 'dynamodb_table = "tfstate_locks"\n'])
    out.close()
    assert filecmp.cmp(filename, compare_file)

def test_gen_terraform_backend_config_s3_access(tmpdir):
    """Testing the generation of a terraform backend config file"""
    config = {'terraform': {'state':{'s3_bucket': 'tf_statefiles',
                                     's3_region': 'us-east-1',
                                     's3_path': 'cb/stage/tf.state',
                                     'dynamodb_table': 'tfstate_locks',
                                     's3_access_key': 'TEST_KEY',
                                     's3_secret_key': 'TEST_SECRET'}}}
    envir = {'TEST_KEY': 'ASASGETXXCD',
             'TEST_SECRET': 'asdfhgbe23d'}
    filename = tmpdir.join('test_summoner_terraform.tf')
    gen_terraform_backend_config(config, filename, envir)

    #Create a config file manually
    compare_file = tmpdir.join('compare.tf')
    out = open(compare_file, "w")
    out.writelines(['region = "us-east-1"\n', 'bucket = "tf_statefiles"\n',
                    'key = "cb/stage/tf.state"\n', 'dynamodb_table = "tfstate_locks"\n',
                    'access_key = "ASASGETXXCD"\n', 'secret_key = "asdfhgbe23d"\n'])
    out.close()
    assert filecmp.cmp(filename, compare_file)

def test_get_env_variables():
    """Testing that environment variables are retrieved correctly """
    config = dict()
    environment = dict()
    config['summoner'] = dict(env_gitlab_api_token='TEST_TOKEN')
    environment['TEST_TOKEN'] = '12345'
    environment['GITLAB_CI'] = 'Test'
    environment['CI_PROJECT_ID'] = 'Test_Project'
    environment['CI_COMMIT_SHA'] = '678910'
    token, project, sha = get_env_variables(config, environment)
    assert token == '12345'
    assert project == 'Test_Project'
    assert sha == '678910'

def test_get_env_variables_config_error():
    """Testing that the ConfigError exception is raised """
    config = dict()
    environment = dict()
    config['summoner'] = dict(env_gitlab_api_token='TEST_TOKEN')
    environment['GITLAB_CI'] = 'Test'
    environment['CI_PROJECT_ID'] = 'Test_Project'
    environment['CI_COMMIT_SHA'] = '678910'
    with pytest.raises(ConfigError) as excinfo:
        get_env_variables(config, environment)
    assert "Unable to load GitLab private token " in str(excinfo.value)

def test_get_env_variables_keyerror_project():
    """
    Testing that the KeyError exception is raised because the
    CI_PROJECT_ID environment variable is missing
    """
    config = dict()
    environment = dict()
    config['summoner'] = dict(env_gitlab_api_token='TEST_TOKEN')
    environment['TEST_TOKEN'] = '12345'
    environment['GITLAB_CI'] = 'Test'
    environment['CI_COMMIT_SHA'] = '678910'
    with pytest.raises(KeyError) as excinfo:
        get_env_variables(config, environment)
    assert "Issue grabbing CI_PROJECT_ID environment variable" in str(excinfo.value)

def test_get_env_variables_keyerror_sha():
    """
    Testing that the KeyError exception is raised because the
    CI_COMMIT_SHA environment variable is missing
    """
    config = dict()
    environment = dict()
    config['summoner'] = dict(env_gitlab_api_token='TEST_TOKEN')
    environment['TEST_TOKEN'] = '12345'
    environment['GITLAB_CI'] = 'Test'
    environment['CI_PROJECT_ID'] = 'Test_Project'
    with pytest.raises(KeyError) as excinfo:
        get_env_variables(config, environment)
    assert "Issue grabbing commit SHA variable" in str(excinfo.value)

def test_backend_config_block_exists_true():
    """
    Testing if the backend configuration block exists in a given
    terraform directory
    """
    # Quick dict to store our various datas.
    data = dict()

    # create our mock result
    data['mock_results'] = dict()
    mock_results = data['mock_results'].copy()

    command = 'grep -nr ' + "'^[[:space:]][[:space:]]backend[[:space:]]\"s3\"[[:space:]]{}'" + ' /terraform/'
    mock_results[command] = [Result("", exited=0)]
    mock_context = MockContext(run=mock_results)

    assert backend_config_block_exists(mock_context, '/terraform') is True

def test_backend_config_block_exists_false():
    """
    Testing if the backend configuration block does not exist in a given
    terraform directory
    """
    # Quick dict to store our various datas.
    data = dict()

    # create our mock result
    data['mock_results'] = dict()
    mock_results = data['mock_results'].copy()

    command = 'grep -nr ' + "'^[[:space:]][[:space:]]backend[[:space:]]\"s3\"[[:space:]]{}'" + ' /terraform/'
    mock_results[command] = [Result("", exited=1)]
    mock_context = MockContext(run=mock_results)

    assert backend_config_block_exists(mock_context, '/terraform') is False
