"""
These test cases are designed to exercise the dry run stage of the
pipeline.
"""
import io

from code.src.summoner.common import ErrorCodes
from code.src.summoner.dryrun import run as dryrun

import mock
from invoke import MockContext, Result
import pytest

# For simplicity, pulled out "success" mocks
# it can be overridden when something needs a failure.

@pytest.fixture(name="fixture_run_data")
def _fixture_run_data():
    """ Some shared data to remove code duplication for tests """
    # Quick dict to store our various datas.
    data = dict()
    # generate config for deploy_targets
    data['config_fd'] = io.StringIO("""
        [summoner]
        env_pagerduty_key_low = "LOW_KEY"
        env_pagerduty_key_high = "HIGH_KEY"

        [[aws-account]]
        account_number = 12347
        deploy_region = 'us-east-1'
        type = 'test'
        env_var_aws_key = 'TEST_KEY'
        env_var_aws_secret = 'TEST_SECRET'

        [[aws-account]]
        account_number = 1234
        deploy_region = 'us-east-1'
        type = 'stage'
        env_var_aws_key = 'STAGE_KEY'
        env_var_aws_secret = 'STAGE_SECRET'
        """)

    data['config_fd2'] = io.StringIO("""
        [summoner]
        env_pagerduty_key_low = "LOW_KEY"
        env_pagerduty_key_high = "HIGH_KEY"

        [[aws-account]]
        account_number = 123478
        deploy_region = 'us-east-1'
        type = 'stage'
        env_var_aws_key = 'STAGE_KEY'
        env_var_aws_secret = 'STAGE_SECRET'
        tfvar_override_files = ["terraform/terraform.tfvars"]
    """)

    # setup the environment:
    data['environment'] = dict(
        LOW_KEY="LOW_KEY",
        HIGH_KEY="HIGH_KEY",
        GL_PRIVATE_TOKEN="GL_PRIVATE_TOKEN",
        CI_PROJECT_PATH_SLUG="test",
        GITLAB_CI="TEST",
        CI_PROJECT_ID="TEST_Project",
        CI_COMMIT_SHA="1234",
        TEST_KEY="abcd",
        TEST_SECRET="1234",
        STAGE_KEY="efgh",
        STAGE_SECRET="5678"
    )

    data['environment2'] = dict(
        TEST_KEY="abcd",
        TEST_SECRET="1234",
        STAGE_KEY="efgh",
        STAGE_SECRET="5678"
    )

    data['mock_results'] = dict()

    return data

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_dryrun_tfinit_fail(fixture_run_data):
    """Testing terraform init"""
    mock_results = fixture_run_data['mock_results'].copy()
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = Result("", exited=1)
    env_var_aws_key = fixture_run_data['environment2']['TEST_KEY']
    env_var_aws_secret = fixture_run_data['environment2']['TEST_SECRET']
    region = "us-east-1"
    mock_results[f"export AWS_ACCESS_KEY_ID={env_var_aws_key} >> ~/.bashrc"] = Result("", exited=0)
    mock_results[f"export AWS_SECRET_ACCESS_KEY={env_var_aws_secret} >> ~/.bashrc"] = Result("", exited=0)
    mock_results[f"export AWS_DEFAULT_REGION={region} >> ~/.bashrc"] = Result("", exited=0)
    assert dryrun(MockContext(run=mock_results),
                  fixture_run_data['config_fd'],
                  environment=fixture_run_data['environment2']) == ErrorCodes.ERROR_TERRAFORM_INIT

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_dryrun_tfapply_fail(fixture_run_data):
    """Testing terraform apply"""
    mock_results = fixture_run_data['mock_results'].copy()
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = Result("", exited=0)
    mock_results['terraform apply -input=false -no-color -auto-approve  terraform'] = Result("", exited=1)
    assert dryrun(MockContext(run=mock_results),
                  fixture_run_data['config_fd'],
                  environment=fixture_run_data['environment2']) == ErrorCodes.ERROR_TERRAFORM_APPLY

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_dryrun_tfdestroy_fail(fixture_run_data):
    """Testing terraform destroy"""
    override_command = ""
    mock_results = fixture_run_data['mock_results'].copy()
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = Result("", exited=0)
    mock_results[f'terraform destroy -input=false -auto-approve {override_command} terraform'] = Result("", exited=1)

    assert dryrun(MockContext(run=mock_results),
                  fixture_run_data['config_fd'],
                  environment=fixture_run_data['environment2'],
                  o_apply_tf=False) == ErrorCodes.ERROR_TERRAFORM_DESTROY

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_apply_success(fixture_run_data):
    """Testing apply success"""
    override_command = ""
    mock_results = fixture_run_data['mock_results'].copy()
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = Result("", exited=0)
    mock_results[f'terraform apply -input=false -no-color -auto-approve {override_command} terraform'] = \
        Result("", exited=0)
    assert dryrun(MockContext(run=mock_results),
                  fixture_run_data['config_fd'],
                  environment=fixture_run_data['environment2'],
                  o_destroy_tf=False) == ErrorCodes.SUCCESS

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_apply_fail_overrides(fixture_run_data):
    """Testing apply fail due to overrides"""
    mock_results = fixture_run_data['mock_results'].copy()
    override_command = " -var-file=terraform/terraform.tfvars"
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = Result(
        "", exited=0)
    mock_results[f'terraform apply -input=false -no-color -auto-approve {override_command} terraform'] = \
        Result("", exited=0)

    assert dryrun(MockContext(run=mock_results),
                  fixture_run_data['config_fd2'],
                  environment=fixture_run_data['environment2'],
                  account_type='stage',
                  o_destroy_tf=False) == ErrorCodes.SUCCESS

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_destroy_success(fixture_run_data):
    """Testing destroy success"""
    override_command = ""
    mock_results = fixture_run_data['mock_results'].copy()
    mock_results[f'terraform destroy -input=false -auto-approve {override_command} terraform'] = Result("", exited=0)
    assert dryrun(MockContext(run=mock_results),
                  fixture_run_data['config_fd'],
                  environment=fixture_run_data['environment2'], o_init_tf=False, o_apply_tf=False) == ErrorCodes.SUCCESS
