"""
These test cases are designed to exercise the launch phase
"""

import io

from code.src.summoner.common import (
    DeployFailure,
    RollbackComplete,
    RollbackFailure,
)

from code.src.summoner.launch import (
    deploy_to_target,
    deploy,
)

import mock
import pytest
from invoke import MockContext, Result


@pytest.fixture(name="fixture_run_data")
def _fixture_run_data():
    """ Some shared data to remove code duplication for tests """
    # Quick dict to store our various datas.
    data = dict()
    # generate config for deploy_targets
    data['config_fd'] = io.StringIO("""
        [summoner]
        env_pagerduty_key_low = "PD_LOW_KEY"
        env_pagerduty_key_high = "PD_HIGH_KEY"

        [[aws-account]]
        account_number = 424242
        deploy_region = 'us-east-1'
        type = 'test'
        env_var_aws_key = 'AWS_TEST_KEY'
        env_var_aws_secret = 'AWS_TEST_SECRET'

        [[aws-account]]
        account_number = 87462
        deploy_region = 'us-east-1'
        type = 'stage'
        env_var_aws_key = 'AWS_STAGE_KEY'
        env_var_aws_secret = 'AWS_STAGE_SECRET'
        """)

    data['config_fd2'] = io.StringIO("""
        [summoner]
        env_pagerduty_key_low = "PD_LOW_KEY"
        env_pagerduty_key_high = "PD_HIGH_KEY"

        [[aws-account]]
        account_number = 887653
        deploy_region = 'us-east-1'
        type = 'stage'
        env_var_aws_key = 'AWS_STAGE_KEY'
        env_var_aws_secret = 'AWS_STAGE_SECRET'
        tfvar_override_files = ["terraform/terraform.tfvars"]
    """)

    # setup the environment:
    data['environment'] = dict(
        PD_LOW_KEY="PD_LOW_KEY",
        PD_HIGH_KEY="PD_HIGH_KEY",
        GL_PRIVATE_TOKEN="GL_PRIVATE_TOKEN",
        CI_PROJECT_PATH_SLUG="test",
        GITLAB_CI="TEST",
        CI_PROJECT_ID="TEST_Project",
        CI_COMMIT_SHA="123424",
        AWS_TEST_KEY="abcdasdfas",
        AWS_TEST_SECRET="1234942",
        AWS_STAGE_KEY="efghasdf",
        AWS_STAGE_SECRET="567888888"
    )

    data['environment2'] = dict(
        AWS_TEST_KEY="sdfnsf",
        AWS_TEST_SECRET="1234321111",
        AWS_STAGE_KEY="efghffvsfvf",
        AWS_STAGE_SECRET="567890709779"
    )

    data['mock_results'] = dict()

    return data

def helper_gen_account(num_files, account_num='0000', deploy_region='test', account_type='test_acct', env_var_aws_key='AWS_TEST_KEY', env_var_aws_secret='AWS_TEST_SECRET'):
    """
    Small helper to generate an account entry for deploy targets
    """
    target_config = dict()
    target_config['account_number'] = account_num
    target_config['deploy_region'] = deploy_region
    target_config['type'] = account_type
    target_config['env_var_aws_key'] = env_var_aws_key
    target_config['env_var_aws_secret'] = env_var_aws_secret

    target_config['tfvar_override_files'] = list()
    for i in range(0, num_files):
        target_config['tfvar_override_files'].append(f'{str(i)}.file')
    return target_config

def helper_noop(*args, **kwargs):
    """
    A simply no-operation function. Useful to dead-end functions
    when running test cases
    """
    print(f"Noop function received: args={args} kwargs={kwargs}")

def helper_gen_deploy_to_target_context(git_exit_code, terraform_exit_code, gitref,
                                        override_command, tf_dir, backend_filepath, fixture_run_data, init_exit_code=0):
    """
    Reduce code duplication by making the context definition generic
    """
    mock_results = dict()
    mock_results[f'git checkout {gitref}'] = \
            Result("", exited=git_exit_code)
    mock_results[f'terraform apply -no-color -auto-approve -input=false {override_command} {tf_dir}'] = \
            Result("", exited=terraform_exit_code)
    mock_results[f'terraform init -reconfigure -backend-config={backend_filepath} -input=false terraform'] = [Result("", exited=init_exit_code)]*2
    env_var_aws_key = fixture_run_data['environment2']['AWS_TEST_KEY']
    env_var_aws_secret = fixture_run_data['environment2']['AWS_TEST_SECRET']
    region = "test"

    mock_results[f"export AWS_ACCESS_KEY_ID={env_var_aws_key} >> ~/.bashrc"] = Result("", exited=0)
    mock_results[f"export AWS_SECRET_ACCESS_KEY={env_var_aws_secret} >> ~/.bashrc"] = Result("", exited=0)
    mock_results[f"export AWS_DEFAULT_REGION={region} >> ~/.bashrc"] = Result("", exited=0)
    return MockContext(run=mock_results)

@mock.patch('code.src.summoner.common.gen_terraform_backend_config')
@mock.patch('code.src.summoner.launch.gen_terraform_backend_config')
def helper_exec_deploy_to_target_success(
        mock_launch_gen_terraform_backend_config,
        mock_common_gen_terraform_backend_config,
        target,
        tf_dir,
        gitref,
        override_command,
        backend_filepath,
        git_exit_code,
        terraform_exit_code,
        environment,
        fixture_run_data):
    """
    This helper is here to remove the boiler-plate code in the
    test functions below. We call the deploy_to_target function
    with whatever context we are provided with.
    """

    mock_context = helper_gen_deploy_to_target_context(
        git_exit_code=git_exit_code,
        terraform_exit_code=terraform_exit_code,
        gitref=gitref,
        override_command=override_command,
        tf_dir=tf_dir,
        backend_filepath=backend_filepath,
        fixture_run_data=fixture_run_data
        )

    deploy_to_target(mock_context, target, backend_filepath, tf_dir, gitref, environment)
    # if something went wrong, we'd get an exception here, and that would fail the test

    # confirm the function invoked the backend config correctly (launch import)
    args, kwargs = mock_launch_gen_terraform_backend_config.call_args
    assert not kwargs.items() # ensure no kwargs passed
    assert args[0] == target
    assert args[1] == backend_filepath

    # confirm the function invoked the backend config correctly (common import)
    args, kwargs = mock_common_gen_terraform_backend_config.call_args
    assert not kwargs.items() # ensure no kwargs passed
    assert args[0] == target
    assert args[1] == backend_filepath

#
# Now for the actual test cases!
#
@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_deploy_to_target_success(fixture_run_data):
    """ When we successfully deploy to a target """
    #pylint: disable=no-value-for-parameter
    # need to disable this pylint check because
    # it thinks we're missing a parameter for this
    # function due to the decorator.
    # https://github.com/PyCQA/pylint/issues/259
    helper_exec_deploy_to_target_success(
        target=helper_gen_account(1),
        tf_dir='terraform',
        gitref='test',
        override_command=" -var-file=0.file",
        backend_filepath='_s3.backend',
        git_exit_code=0,
        terraform_exit_code=0,
        environment=fixture_run_data["environment2"],
        fixture_run_data=fixture_run_data)

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_deploy_to_target_git_none(fixture_run_data):
    """
    Here we test that a gitref of None correctly skips
    the git checkout call. We confirm this by instructing
    invoke to return a non-0 exit code, which would result in an
    exception if that code block is called
    """
    #pylint: disable=no-value-for-parameter
    # need to disable this pylint check because
    # it thinks we're missing a parameter for this
    # function due to the decorator.
    # https://github.com/PyCQA/pylint/issues/259
    helper_exec_deploy_to_target_success(
        target=helper_gen_account(1),
        tf_dir='terraform',
        gitref=None,
        override_command=" -var-file=0.file",
        backend_filepath="_s3.backend",
        git_exit_code=1,
        terraform_exit_code=0,
        environment=fixture_run_data["environment2"],
        fixture_run_data=fixture_run_data)

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_deploy_to_target_zero_override(fixture_run_data):
    """
    Test that zero overrides are properly passed to terraform.
    """
    #pylint: disable=no-value-for-parameter
    # need to disable this pylint check because
    # it thinks we're missing a parameter for this
    # function due to the decorator.
    # https://github.com/PyCQA/pylint/issues/259
    helper_exec_deploy_to_target_success(
        target=helper_gen_account(0),
        tf_dir='terraform',
        gitref=None,
        override_command="",
        backend_filepath="_s3.backend",
        git_exit_code=1,
        terraform_exit_code=0,
        environment=fixture_run_data["environment2"],
        fixture_run_data=fixture_run_data)

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.common.gen_terraform_backend_config', mock.Mock())
def test_deploy_to_target_three_overrides(fixture_run_data):
    """
    Test that three overrides are properly passed to terraform.
    """
    #pylint: disable=no-value-for-parameter
    # need to disable this pylint check because
    # it thinks we're missing a parameter for this
    # function due to the decorator.
    # https://github.com/PyCQA/pylint/issues/259
    helper_exec_deploy_to_target_success(
        target=helper_gen_account(3),
        tf_dir='terraform',
        gitref=None,
        backend_filepath="_s3.backend",
        override_command=" -var-file=0.file -var-file=1.file -var-file=2.file",
        git_exit_code=1,
        terraform_exit_code=0,
        environment=fixture_run_data["environment2"],
        fixture_run_data=fixture_run_data)

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.launch.gen_terraform_backend_config', mock.Mock())
@mock.patch('code.src.summoner.common.gen_terraform_backend_config')
def test_deploy_to_target_git_failure(mock_gen_terraform_backend_config, fixture_run_data):
    """
    Test that a git failure is correctly raised as a ValueError
    """

    mock_gen_terraform_backend_config.side_effect = helper_noop
    target = helper_gen_account(1)
    tf_dir = 'terraform'
    gitref = 'test'
    backend_filepath = "_s3.backend"

    mock_context = helper_gen_deploy_to_target_context(
        git_exit_code=1,
        terraform_exit_code=0,
        gitref=gitref,
        backend_filepath=backend_filepath,
        override_command=" -var-file=0.file",
        tf_dir=tf_dir,
        fixture_run_data=fixture_run_data)

    with pytest.raises(ValueError) as excinfo:
        deploy_to_target(mock_context, target, backend_filepath, tf_dir, gitref, fixture_run_data["environment2"])
    assert "Unable to checkout git reference" in str(excinfo.value)

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.launch.gen_terraform_backend_config', mock.Mock())
@mock.patch('code.src.summoner.common.gen_terraform_backend_config')
def test_deploy_to_target_terraform_failure(mock_gen_terraform_backend_config, fixture_run_data):
    """
    Test that a failure in the terraform apply is reported
    as a DeployFailure exception
    """

    mock_gen_terraform_backend_config.side_effect = helper_noop
    target = helper_gen_account(1)
    tf_dir = 'terraform'
    gitref = 'test'
    backend_filepath = '_s3.backend'

    mock_context = helper_gen_deploy_to_target_context(
        git_exit_code=0,
        terraform_exit_code=1,
        gitref=gitref,
        backend_filepath=backend_filepath,
        override_command=" -var-file=0.file",
        tf_dir=tf_dir,
        fixture_run_data=fixture_run_data)

    with pytest.raises(DeployFailure) as excinfo:
        deploy_to_target(mock_context, target, backend_filepath, tf_dir, gitref, fixture_run_data["environment2"])
    assert "Error applying terraform" in str(excinfo.value)

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.launch.gen_terraform_backend_config', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.common.gen_terraform_backend_config')
def test_deploy_to_target_terraform_init_failure(mock_gen_terraform_backend_config, fixture_run_data):
    """
    Test that a failure in the terraform init is reported
    as a DeployFailure exception
    """

    mock_gen_terraform_backend_config.side_effect = helper_noop
    target = helper_gen_account(1)
    tf_dir = 'terraform'
    gitref = 'test'

    mock_context = helper_gen_deploy_to_target_context(
        git_exit_code=0,
        terraform_exit_code=0,
        init_exit_code=1,
        gitref=gitref,
        override_command=" -var-file=0.file",
        backend_filepath='_s3.backend',
        tf_dir=tf_dir,
        fixture_run_data=fixture_run_data)

    with pytest.raises(DeployFailure) as excinfo:
        deploy_to_target(mock_context, target, "_s3.backend", tf_dir, gitref, fixture_run_data["environment2"])
    assert "Error initializing terraform" in str(excinfo.value)


@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.launch.find_last_deploy')
@mock.patch('code.src.summoner.launch.deploy_to_target')
def test_deploy_success(mock_deploy_to_target, mock_find_last_deploy, fixture_run_data):
    """ Run deploy and confirm success. Success here == no exceptions. """
    account_type = "type1"

    targets = dict()
    targets[account_type] = list()
    targets[account_type].append(helper_gen_account(1, account_num='0000', deploy_region='test'))

    context = MockContext()
    deploy_targets = targets
    tf_dir = "terraform"
    git_tag_prefix = f"summoner-{account_type}-r"
    o_rollback_disabled = False
    mock_find_last_deploy.side_effect = [2]  # return a reasonable rollback version

    deploy(context, account_type, deploy_targets, "_s3.backend", tf_dir, git_tag_prefix, fixture_run_data["environment2"], o_rollback_disabled)

    # confirm our mock functions were called the correct number of times.
    assert len(mock_deploy_to_target.call_args_list) == 1
    assert len(mock_find_last_deploy.call_args_list) == 1

    # Check the correct arguments were passed on first invocation
    args, kwargs = mock_deploy_to_target.call_args_list[0]
    assert not kwargs.items() # ensure no kwargs passed
    assert args[0] == context
    assert args[1] == targets[account_type][0]
    assert args[3] == '_s3.backend'
    assert args[2] == tf_dir
    assert args[4] is None

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.launch.find_last_deploy')
@mock.patch('code.src.summoner.launch.deploy_to_target')
def test_deploy_fail_rollback_success(mock_deploy_to_target, mock_find_last_deploy, fixture_run_data):
    """ Run deploy and confirm successful rollback. Success here == no exceptions. """
    account_type = "type1"

    targets = dict()
    targets[account_type] = list()
    targets[account_type].append(helper_gen_account(1, account_num='0000', deploy_region='test'))

    context = MockContext()
    deploy_targets = targets
    tf_dir = "terraform"
    git_tag_prefix = f"summoner-{account_type}-r"
    git_tag_rollback_version = 2
    o_rollback_disabled = False
    mock_find_last_deploy.side_effect = [git_tag_rollback_version]  # return a reasonable rollback version

    # in order to raise an exception to cause a rollback, we raise
    # the correct exception to begin with, then return no exception
    # the next time.
    mock_deploy_to_target.side_effect = [
        DeployFailure("Test Failure"),
        None
    ]

    with pytest.raises(RollbackComplete) as excinfo:
        deploy(context, account_type, deploy_targets, "_s3.backend", tf_dir, git_tag_prefix, fixture_run_data["environment2"], o_rollback_disabled)
    assert f"Deploy failed for {account_type}:0000. Rollback" in str(excinfo.value)

    # confirm our mock functions were called the correct number of times.
    assert len(mock_deploy_to_target.call_args_list) == 2
    assert len(mock_find_last_deploy.call_args_list) == 1

    # Check the correct arguments were passed on first invocation
    args, kwargs = mock_deploy_to_target.call_args_list[0]
    assert not kwargs.items() # ensure no kwargs passed
    assert args[0] == context
    assert args[1] == targets[account_type][0]
    assert args[2] == tf_dir
    assert args[3] == '_s3.backend'
    assert args[4] is None

    # Check the correct rollback arguments were passed on second invocation
    args, kwargs = mock_deploy_to_target.call_args_list[1]
    assert not kwargs.items() # ensure no kwargs passed
    assert args[0] == context
    assert args[1] == targets[account_type][0]
    assert args[2] == tf_dir
    assert args[3] == '_s3.backend'
    assert args[4] == f"summoner-{account_type}-r{git_tag_rollback_version}"

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.launch.find_last_deploy')
@mock.patch('code.src.summoner.launch.deploy_to_target')
def test_deploy_fail_rollback_fail(mock_deploy_to_target, mock_find_last_deploy, fixture_run_data):
    """ Run deploy and confirm failed rollback is raised as exception """
    account_type = "type1"

    targets = dict()
    targets[account_type] = list()
    targets[account_type].append(helper_gen_account(1, account_num='0000', deploy_region='test'))

    context = MockContext()
    deploy_targets = targets
    tf_dir = "terraform"
    git_tag_prefix = f"summoner-{account_type}-r"
    git_tag_rollback_version = 2
    o_rollback_disabled = False
    mock_find_last_deploy.side_effect = [git_tag_rollback_version]  # return a reasonable rollback version

    # in order to raise an exception to cause a rollback, we raise
    # the correct exception to begin with, then return no exception
    # the next time.
    mock_deploy_to_target.side_effect = [
        DeployFailure("Test Failure"),
        DeployFailure("Test Rollback Failure")
    ]

    with pytest.raises(RollbackFailure) as excinfo:
        deploy(context, account_type, deploy_targets, tf_dir, "_s3.backend", git_tag_prefix, fixture_run_data["environment2"], o_rollback_disabled)
    assert "Test Rollback Failure" in str(excinfo.value)

    # confirm our mock functions were called the correct number of times.
    assert len(mock_deploy_to_target.call_args_list) == 2
    assert len(mock_find_last_deploy.call_args_list) == 1

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.launch.find_last_deploy')
@mock.patch('code.src.summoner.launch.deploy_to_target')
def test_deploy_fail_no_rollback(mock_deploy_to_target, mock_find_last_deploy, fixture_run_data):
    """ Run deploy fail, and then no rollback because no git ref """
    account_type = "type1"

    targets = dict()
    targets[account_type] = list()
    targets[account_type].append(helper_gen_account(1, account_num='0000', deploy_region='test'))

    context = MockContext()
    deploy_targets = targets
    tf_dir = "terraform"
    git_tag_prefix = f"summoner-{account_type}-r"
    git_tag_rollback_version = None
    o_rollback_disabled = False
    mock_find_last_deploy.side_effect = [git_tag_rollback_version]  # return a reasonable rollback version

    # in order to raise an exception to cause a rollback, we raise
    # the correct exception to begin with, then return no exception
    # the next time.
    mock_deploy_to_target.side_effect = [
        DeployFailure("Test Failure"),
        None
    ]

    with pytest.raises(DeployFailure) as excinfo:
        deploy(context, account_type, deploy_targets, tf_dir, "_s3.backend", git_tag_prefix, fixture_run_data["environment2"], o_rollback_disabled)
    assert "Deploy failed and no rollback" in str(excinfo.value)

    # confirm our mock functions were called the correct number of times.
    assert len(mock_deploy_to_target.call_args_list) == 1
    assert len(mock_find_last_deploy.call_args_list) == 1

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.launch.find_last_deploy')
@mock.patch('code.src.summoner.launch.deploy_to_target')
def test_deploy_multiple_deploy_fail_rollback_success(mock_deploy_to_target, mock_find_last_deploy, fixture_run_data):
    """ Run deploy and confirm failed rollback is raised as exception """
    account_type = "type1"

    targets = dict()
    targets[account_type] = list()
    targets[account_type].append(helper_gen_account(1, account_num='0000', deploy_region='test'))
    targets[account_type].append(helper_gen_account(1, account_num='0001', deploy_region='test'))

    context = MockContext()
    deploy_targets = targets
    tf_dir = "terraform"
    git_tag_prefix = f"summoner-{account_type}-r"
    git_tag_rollback_version = 2
    o_rollback_disabled = False
    mock_find_last_deploy.side_effect = [git_tag_rollback_version]  # return a reasonable rollback version

    # in order to raise an exception to cause a rollback, we raise
    # the correct exception to begin with, then return no exception
    # the next time.
    mock_deploy_to_target.side_effect = [
        None,  # deploy first account
        DeployFailure("Test Failure"),  # fail deploy of second account
        None,  # rollback second account
        None,  # rollback first account
    ]

    with pytest.raises(RollbackComplete) as excinfo:
        deploy(context, account_type, deploy_targets, tf_dir, "_s3.backend", git_tag_prefix, fixture_run_data["environment2"], o_rollback_disabled)
    assert f"Deploy failed for {account_type}:0001. Rollback" in str(excinfo.value)

    # confirm our mock functions were called the correct number of times.
    assert len(mock_deploy_to_target.call_args_list) == 4
    assert len(mock_find_last_deploy.call_args_list) == 1

    # confirm the deploy/rollback order was correct
    args = mock_deploy_to_target.call_args_list[0][0]  # get args
    assert args[1] == targets[account_type][0]  # first account deploy
    args = mock_deploy_to_target.call_args_list[1][0]  # get args
    assert args[1] == targets[account_type][1]  # second account deploy
    args = mock_deploy_to_target.call_args_list[2][0]  # get args
    assert args[1] == targets[account_type][1]  # second account rollback
    args = mock_deploy_to_target.call_args_list[3][0]  # get args
    assert args[1] == targets[account_type][0]  # first account rollback
