"""
These test cases are designed to exercise the launch phase

The goal for these test cases is to ensure the whole launch phase
is knitted together correctly (function calls correct, etc.) As such,
we allow all helper functions to execute, and we mock as few of the
results as possible to get as close to true functionality as possible.

"""

import io

from code.src.summoner.launch import run
from code.src.summoner.common import ErrorCodes

from invoke import MockContext, Result
import mock
import pytest
import gitlab

@pytest.fixture(name="fixture_run_data")
def _fixture_run_data():
    """ Some shared data to remove code duplication for tests """

    # Quick dict to store our various datas.
    data = dict()

    # mock the pagerduty response
    data['mock_pypd_eventv2_create_return_value'] = {
        "status": "success",
        "message": "Event processed",
        "dedup_key": "01234"
    }

    # generate config for deploy_targets
    data['config_fd'] = io.StringIO("""
        [summoner]
        env_pagerduty_key_low = "PD_LOW_KEY"
        env_pagerduty_key_high = "PD_HIGH_KEY"
        env_gitlab_api_token = "GL_PRIVATE_TOKEN"

        [[aws-account]]
        account_number = 123
        deploy_region = 'us-east-1'
        type = 'stage'
        env_var_aws_key = 'AWS_STAGE_KEY'
        env_var_aws_secret = 'AWS_STAGE_SECRET'


        [[aws-account]]
        account_number = 1234
        deploy_region = 'us-east-1'
        type = 'prod'
        env_var_aws_key = 'AWS_PROD_KEY'
        env_var_aws_secret = 'AWS_PROD_SECRET'

        """)

    # setup the git ref return:
    data['tag_list'] = [
        '2262382e685462c78edc48de81da1daacd6fe679 refs/tags/summoner-stage-1',
        '2262382e685462c78edc48de81da1daacd6fe679 refs/tags/summoner-prod-r11',
        '2262382e685462c78edc48de81da1daacd6fe679 refs/tags/summoner-stage-r2',
        '2262382e685462c78edc48de81da1daacd6fe679 refs/tags/summoner-prod-r3',
        '2262382e685462c78edc48de81da1daacd6fe679 refs/tags/summoner-stage-r30'
    ]
    data['tag_blob'] = "\n".join(data['tag_list'])

    # setup the environment:
    data['environment'] = dict(
        PD_LOW_KEY="PD_LOW_KEY",
        PD_HIGH_KEY="PD_HIGH_KEY",
        GL_PRIVATE_TOKEN="GL_PRIVATE_TOKEN",
        CI_PROJECT_PATH_SLUG="test",
        GITLAB_CI="TEST",
        CI_PROJECT_ID="TEST_Project",
        CI_COMMIT_SHA="1234",
        AWS_PROD_KEY="abcdtdg",
        AWS_PROD_SECRET="12345345",
        AWS_STAGE_KEY="efghsdfgg",
        AWS_STAGE_SECRET="567823456"
    )

    # create our mock result
    data['mock_results'] = dict()
    data['mock_results']["git show-ref --tags"] = \
            [Result(data['tag_blob'], exited=0)] * 4
    data['mock_results']['git checkout summoner-prod-r11'] = \
            Result("", exited=0)
    return data


@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.launch.gitlab_push_tag', mock.Mock())
@mock.patch('pypd.EventV2.create')
def test_launch_run_rollback_complete(mock_pypd_eventv2_create, fixture_run_data):
    """
    This test will test
      1)  deploy code to stage (expect success)
      2)  deploy code to prod (expect fail)
      3)  rollback code in prod (expect success)
    """
    mock_pypd_eventv2_create.return_value = fixture_run_data['mock_pypd_eventv2_create_return_value']


    mock_results = fixture_run_data['mock_results'].copy()
    mock_results["git tag summoner-stage-r31"] = \
            Result("", exited=0)
    mock_results["git push origin summoner-stage-r31"] = \
            Result("", exited=0)
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = [Result("", exited=0)]*3
    mock_results['terraform apply -no-color -auto-approve -input=false  terraform'] = [
        Result("", exited=0),  # stage deploy works
        Result("", exited=1),  # prod deploy fails
        Result("", exited=0)   # prod rollback works
    ]
    mock_context = MockContext(run=mock_results)
    # Run the phase, see what happens!
    return_code = run(
        mock_context,
        fixture_run_data['config_fd'],
        o_launch_to_prod=True,
        environment=fixture_run_data["environment"]
    )
    assert return_code == ErrorCodes.ERROR_LAUNCH_RECOVERY


@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.launch.gitlab_push_tag', mock.Mock())
@mock.patch('pypd.EventV2.create')
def test_launch_run_rollback_halt(mock_pypd_eventv2_create, fixture_run_data):
    """
    This test will test
      1)  deploy code to stage (expect success)
      2)  deploy code to prod, no rollback (expect fail)
    """
    mock_pypd_eventv2_create.return_value = fixture_run_data['mock_pypd_eventv2_create_return_value']


    mock_results = fixture_run_data['mock_results'].copy()
    mock_results["git tag summoner-stage-r31"] = \
            Result("", exited=0)
    mock_results["git push origin summoner-stage-r31"] = \
            Result("", exited=0)
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = [Result("", exited=0)]*2
    mock_results['terraform apply -no-color -auto-approve -input=false  terraform'] = [
        Result("", exited=0),  # stage deploy works
        Result("", exited=1),  # prod deploy fails, no rollback
    ]
    mock_context = MockContext(run=mock_results)
    # Run the phase, see what happens!
    return_code = run(
        mock_context,
        fixture_run_data['config_fd'],
        o_launch_to_prod=True,
        environment=fixture_run_data["environment"],
        o_rollback_disabled=True
    )
    assert return_code == ErrorCodes.ERROR_TERRAFORM_APPLY


@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.launch.gitlab_push_tag', mock.Mock())
@mock.patch('pypd.EventV2.create')
def test_launch_run_rollback_halt_staging(mock_pypd_eventv2_create, fixture_run_data):
    """
    This test will test
      1)  deploy code to stage (expect success)
      2)  deploy code to prod, no rollback (expect fail)
    """
    mock_pypd_eventv2_create.return_value = fixture_run_data['mock_pypd_eventv2_create_return_value']


    mock_results = fixture_run_data['mock_results'].copy()
    mock_results["git tag summoner-stage-r31"] = \
            Result("", exited=0)
    mock_results["git push origin summoner-stage-r31"] = \
            Result("", exited=0)
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = [Result("", exited=0)]
    mock_results['terraform apply -no-color -auto-approve -input=false  terraform'] = [
        Result("", exited=1),  # stage deploy fails, no rollback
        ]
    mock_context = MockContext(run=mock_results)
    # Run the phase, see what happens!
    return_code = run(
        mock_context,
        fixture_run_data['config_fd'],
        o_launch_to_prod=True,
        environment=fixture_run_data["environment"],
        o_rollback_disabled=True
    )
    assert return_code == ErrorCodes.ERROR_TERRAFORM_APPLY



@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.launch.gitlab_push_tag', mock.Mock())
@mock.patch('code.src.summoner.launch.gen_terraform_backend_config')
@mock.patch('pypd.EventV2.create')
def test_launch_run_success(mock_pypd_eventv2_create, mock_gen_backend_config, fixture_run_data):
    """
    This test will test
      1)  deploy code to stage (expect success)
      2)  deploy code to prod (expect success)
    """
    mock_pypd_eventv2_create.return_value = fixture_run_data['mock_pypd_eventv2_create_return_value']

    mock_results = fixture_run_data['mock_results'].copy()
    mock_results["git tag summoner-stage-r31"] = \
            Result("", exited=0)
    mock_results["git tag summoner-prod-r31"] = \
            Result("", exited=0)
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = [Result("", exited=0)]*2
    mock_results['terraform apply -no-color -auto-approve -input=false  terraform'] = [
        Result("", exited=0),  # stage deploy works
        Result("", exited=0),  # prod deploy works
    ]

    mock_context = MockContext(run=mock_results)

    # Run the phase, see what happens!
    return_code = run(
        mock_context,
        fixture_run_data['config_fd'],
        backend_filepath='_s3_backend.summoner',
        o_launch_to_prod=True,
        environment=fixture_run_data['environment']
    )

    # confirm we generate the correct filename for the backend config:
    backend_config_args = mock_gen_backend_config.call_args_list
    # unpacking [list of calls](arg from arg,kwarg tuple)[ 1st item in arg list ]
    assert backend_config_args[0][0][1] == '_s3_backend.summoner'
    assert backend_config_args[1][0][1] == '_s3_backend.summoner'

    assert return_code == ErrorCodes.SUCCESS

@mock.patch('pypd.EventV2.create')
def test_launch_run_rollback_fail(mock_pypd_eventv2_create, fixture_run_data):
    """
    This test will test
      1)  deploy code to stage (expect success)
      2)  deploy code to prod (expect failure)
      3)  rollback code in prod (expect failure)
    """
    mock_pypd_eventv2_create.return_value = fixture_run_data['mock_pypd_eventv2_create_return_value']

    mock_results = fixture_run_data['mock_results'].copy()
    mock_results["git tag summoner-stage-r31"] = \
            Result("", exited=0)
    mock_results['terraform apply -no-color -auto-approve -input=false  terraform'] = [
        Result("", exited=0),  # stage deploy works
        Result("", exited=1),  # prod deploy fails
        Result("", exited=1),  # prod rollback fails
    ]
    mock_context = MockContext(run=mock_results)

    # Run the phase, see what happens!
    return_code = run(
        mock_context,
        fixture_run_data['config_fd'],
        o_launch_to_prod=True,
        environment=fixture_run_data['environment']
    )

    assert len(mock_pypd_eventv2_create.call_args_list) == 1
    assert return_code == ErrorCodes.ERROR_LAUNCH_FAILURE

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('pypd.EventV2.create')
def test_launch_run_git_tag_fail(mock_pypd_eventv2_create, fixture_run_data):
    """
    This test will test
      1)  deploy code to stage (expect success)
      2)  when using git tag, will fail
    """
    mock_pypd_eventv2_create.return_value = fixture_run_data['mock_pypd_eventv2_create_return_value']

    mock_results = fixture_run_data['mock_results'].copy()
    mock_results["git tag summoner-stage-r31"] = \
            Result("", exited=1)  # fail the git tag
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = Result("", exited=0)
    mock_results['terraform apply -no-color -auto-approve -input=false  terraform'] = [
        Result("", exited=0),  # stage deploy works
    ]

    mock_context = MockContext(run=mock_results)

    with pytest.raises(Exception) as excinfo:
        # Run the phase, see what happens!
        run(
            mock_context,
            fixture_run_data['config_fd'],
            o_launch_to_prod=True,
            environment=fixture_run_data['environment']
        )
    assert "Unable to tag " in str(excinfo.value)

def auth_fail(*args, **kwargs):
    # pylint: disable=unused-argument
    """Testing Gitlab Auth phase"""
    raise gitlab.GitlabAuthenticationError("Bad authentication, check your Gitlab Private Key")

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.launch.gitlab_push_tag')
@mock.patch('pypd.EventV2.create')
def test_gl_tag_push_auth_failure(mock_pypd_eventv2_create, mock_gitlab_push_tag, fixture_run_data):
    """Test Gitlab Auth phase"""
    mock_pypd_eventv2_create.return_value = fixture_run_data['mock_pypd_eventv2_create_return_value']
    mock_gitlab_push_tag.side_effect = auth_fail
    mock_results = fixture_run_data['mock_results'].copy()
    mock_results["git tag summoner-stage-r31"] = \
            Result("", exited=0)
    mock_results["git tag summoner-prod-r31"] = \
            Result("", exited=0)
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = [Result("", exited=0)]*2
    mock_results['terraform apply -no-color -auto-approve -input=false  terraform'] = [
        Result("", exited=0),  # stage deploy works
        Result("", exited=0),  # prod deploy works
    ]

    mock_context = MockContext(run=mock_results)
      # Run the phase, see what happens!
    assert run(mock_context,
               fixture_run_data['config_fd'],
               o_launch_to_prod=True,
               environment=fixture_run_data['environment']) == ErrorCodes.ERROR_GITLAB


def test_launch_run_bad_pagerduty_keys_short(fixture_run_data):
    """ If short (or nonexistant) pagerduty keys are specified
    raise an error. This particular edge case caused me much problems
    before I could figure this out.  Here's a test case for it so
    we don't have future headaches. """
    with pytest.raises(ValueError) as excinfo:
        run(
            MockContext(),
            fixture_run_data['config_fd'],
            o_launch_to_stage=False,
            o_launch_to_prod=False,
            environment={"PD_LOW_KEY": 'bad',
                         "PD_HIGH_KEY": 'bad',
                         "GL_PRIVATE_TOKEN": 'GL_PRIVATE_TOKEN'}
        )
    assert "PagerDuty keys are too short" in str(excinfo.value)

def test_launch_run_bad_pagerduty_keys_nonexistant(fixture_run_data):
    """ If the pagerduty keys don't exist, an error should be raised """
    with pytest.raises(KeyError) as excinfo:
        run(
            MockContext(),
            fixture_run_data['config_fd'],
            o_launch_to_stage=False,
            o_launch_to_prod=False,
            environment={"GL_PRIVATE_TOKEN": 'GL_PRIVATE_TOKEN'}
        )
    assert "Unable to read PagerDuty keys" in str(excinfo.value)
