"""
These test cases are designed to exercise the plan stage of the
pipeline.
"""

from copy import deepcopy
import io

from code.src.summoner.common import ErrorCodes
from code.src.summoner.plan import run as plan

import gitlab
import mock
import pytest
from invoke import MockContext, Result

# generate config for targets
@pytest.fixture(name="val_phase_mock_data")
def _val_phase_mock_data():
    data = dict()

    data['config_fd'] = io.StringIO("""
        [summoner]
        env_gitlab_api_token = "GL_TEST_TOKEN"

        [[aws-account]]
        account_number = 789
        deploy_region = 'summoner-plan-1'
        type = 'stage'
        env_var_aws_key = 'AWS_STAGE_KEY'
        env_var_aws_secret = 'AWS_STAGE_SECRET'

        [[aws-account]]
        account_number = 6789
        deploy_region = 'summoner-plan-1'
        type = 'prod'
        env_var_aws_key = 'AWS_PROD_KEY'
        env_var_aws_secret = 'AWS_PROD_SECRET'
        """)

    data['config_fd_2'] = io.StringIO("""
        [summoner]
        env_gitlab_api_token = "GL_TEST_TOKEN"

        [[aws-account]]
        account_number = 789
        deploy_region = 'summoner-plan-1'
        type = 'stage'
        env_var_aws_key = 'AWS_STAGE_KEY'
        env_var_aws_secret = 'AWS_STAGE_SECRET'

        [[aws-account]]
        account_number = 789
        deploy_region = 'summoner-plan-2'
        type = 'stage'
        env_var_aws_key = 'AWS_STAGE_KEY'
        env_var_aws_secret = 'AWS_STAGE_SECRET'

        [[aws-account]]
        account_number = 6789
        deploy_region = 'summoner-plan-1'
        type = 'prod'
        env_var_aws_key = 'AWS_PROD_KEY'
        env_var_aws_secret = 'AWS_PROD_SECRET'
        """)

    data['config_fd_overrides'] = io.StringIO("""
        [summoner]
        env_gitlab_api_token = "GL_TEST_TOKEN"

        [[aws-account]]
        account_number = 789
        deploy_region = 'summoner-plan-1'
        type = 'stage'
        env_var_aws_key = 'AWS_STAGE_KEY'
        env_var_aws_secret = 'AWS_STAGE_SECRET'
        tfvar_override_files = ['one.tfvar', 'two.tfvar']
        """)

    # setup the environment:
    data['environment'] = dict(
        PD_LOW_KEY="PD_LOW_KEY",
        PD_HIGH_KEY="PD_HIGH_KEY",
        GL_PRIVATE_TOKEN="GL_PRIVATE_TOKEN",
        CI_PROJECT_PATH_SLUG="test",
        GITLAB_CI="TEST",
        CI_PROJECT_ID="TEST_Project",
        CI_COMMIT_SHA="1234",
        AWS_TEST_KEY="abcdfasdf",
        AWS_TEST_SECRET="1234432",
        AWS_STAGE_KEY="efghasdf",
        AWS_STAGE_SECRET="567832",
        AWS_PROD_KEY="efghwdfq",
        AWS_PROD_SECRET="56782323"
    )
    # A generic env variable for ease of testing below
    data['env'] = {
        'GL_TEST_TOKEN': "test_token",
        'AWS_STAGE_KEY': "efghasdf",
        'AWS_STAGE_SECRET': "4523452",
        'AWS_PROD_KEY':"efghwdfq",
        'AWS_PROD_SECRET':"12324"
    }

    return data

# For simplicity, pulled out "success" mocks
# it can be overridden when something needs a failure.
VALIDATE_MOCK_RESULTS = {
    'terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform': Result(""),
    'terraform plan -no-color  terraform': Result("")
}

def test_validate_gitlab_token_fail(val_phase_mock_data):
    """ Confirm a key error is raised when the GL token is missing """
    assert plan(MockContext(), config_file=val_phase_mock_data['config_fd'],
                environment={}) == ErrorCodes.ERROR_MISSING_CONFIG

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_validate_tfinit_fail(val_phase_mock_data):
    """Testing stub functions"""
    mock_results = deepcopy(VALIDATE_MOCK_RESULTS)
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = Result("", exited=1)

    assert plan(MockContext(run=mock_results), config_file=val_phase_mock_data['config_fd'],
                environment=val_phase_mock_data['env']) == ErrorCodes.ERROR_TERRAFORM_INIT

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_validate_tfplan_fail(val_phase_mock_data):
    """Testing stub functions"""
    mock_results = deepcopy(VALIDATE_MOCK_RESULTS)
    mock_results['terraform plan -no-color  terraform'] = Result("", exited=1)
    assert plan(MockContext(run=mock_results), config_file=val_phase_mock_data['config_fd'],
                environment=val_phase_mock_data['env']) == ErrorCodes.ERROR_TERRAFORM_PLAN

def test_set_env_variables_fail(val_phase_mock_data):
    """Testing setting environment variables"""
    environment = val_phase_mock_data['env']
    environment['GITLAB_CI'] = "TEST"
    assert plan(MockContext(run=VALIDATE_MOCK_RESULTS), config_file=val_phase_mock_data['config_fd'],
                environment=environment) == ErrorCodes.ERROR_MISSING_PARAMS

def test_commit_sha_env_fail(val_phase_mock_data):
    """Testing failure to retrieve commit SHA"""
    environment = val_phase_mock_data['env']
    environment.update({'GITLAB_CI': "TEST", 'CI_PROJECT_ID': "TEST"})
    assert plan(MockContext(run=VALIDATE_MOCK_RESULTS), config_file=val_phase_mock_data['config_fd'],
                environment=environment) == ErrorCodes.ERROR_MISSING_PARAMS

def auth_fail(**kwargs):
    # pylint: disable=unused-argument
    """Testing Gitlab Auth phase"""
    raise gitlab.GitlabAuthenticationError("Bad authentication, check your Gitlab Private Key")

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch('code.src.summoner.plan.gitlab_comment_on_commit')
def test_gitlab_auth_failure(mock_comment_on_commit, val_phase_mock_data):
    """Test Gitlab Auth phase"""
    environment = val_phase_mock_data['env']
    environment.update({'GITLAB_CI': "TEST",
                        'CI_PROJECT_ID': "TEST",
                        'CI_COMMIT_SHA': "TEST"})
    mock_comment_on_commit.side_effect = auth_fail
    mock_results = deepcopy(VALIDATE_MOCK_RESULTS)

    assert plan(MockContext(run=mock_results), config_file=val_phase_mock_data['config_fd'],
                environment=environment) == ErrorCodes.ERROR_GITLAB

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_plan_success(val_phase_mock_data):
    """Testing stub functions"""
    environment = val_phase_mock_data['env']
    mock_results = deepcopy(VALIDATE_MOCK_RESULTS)
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = Result("", exited=0)
    mock_results['terraform plan -no-color  terraform'] = Result("", exited=0)

    assert plan(MockContext(run=mock_results), config_file=val_phase_mock_data['config_fd'],
                environment=environment) == ErrorCodes.SUCCESS

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_plan_success_overrides(val_phase_mock_data):
    """Testing stub functions"""
    environment = val_phase_mock_data['env']
    mock_results = deepcopy(VALIDATE_MOCK_RESULTS)
    mock_results['terraform plan -no-color -var-file=one.tfvar -var-file=two.tfvar terraform'] = Result("", exited=0)

    assert plan(MockContext(run=mock_results), config_file=val_phase_mock_data['config_fd_overrides'],
                environment=environment) == ErrorCodes.SUCCESS


@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
@mock.patch("code.src.summoner.common.gen_terraform_backend_config", mock.Mock())
@mock.patch("code.src.summoner.plan.gen_terraform_backend_config")
def test_plan_multiaccount_success(mock_plan_gen_backend_config, val_phase_mock_data):
    """ Confirm we init two accounts correctly """
    mock_results = deepcopy(VALIDATE_MOCK_RESULTS)
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = [Result("", exited=0)]*2
    mock_results['terraform plan -no-color  terraform'] = [Result("", exited=0)]*2

    assert plan(MockContext(run=mock_results), config_file=val_phase_mock_data['config_fd_2'],
                environment=val_phase_mock_data['env']) == ErrorCodes.SUCCESS

    assert len(mock_plan_gen_backend_config.call_args_list) == 2  # ensure we were called twice
