"""
These test cases are designed to exercise the validation stage of the
pipeline.
"""

import io

from code.src.summoner.common import ErrorCodes
from code.src.summoner.validate import run

import mock
from invoke import MockContext, Result
import pytest

@pytest.fixture(name="fixture_run_data")
def _fixture_run_data():
    """ Some shared data to remove code duplication for tests """
    #pylint: disable=duplicate-code
    # Quick dict to store our various datas.
    data = dict()

    # Generate config for deploy_targets
    data['config_fd'] = io.StringIO("""
        [summoner]
        env_pagerduty_key_low = "PD_LOW_KEY"
        env_pagerduty_key_high = "PD_HIGH_KEY"
        env_gitlab_api_token = "GL_PRIVATE_TOKEN"

        [[aws-account]]
        account_number = 1234567
        deploy_region = 'us-east-1'
        type = 'test'
        env_var_aws_key = 'AWS_TEST_KEY'
        env_var_aws_secret = 'AWS_TEST_SECRET'
            [aws-account.terraform.state]
          # An optional bracket. This will override the default
          # s3 bucket and path generated for maintaining state.
          s3_region = "us-east-1"				# region of the S3 bucket
          s3_bucket = "cb_shared_services_tf_statefiles"        # S3 bucket holding the terraform state file
          s3_path = "/exampleproject/test/tf.state"             # Path in the S3 bucket where the terraform state file lives
          dynamodb_table = "cb_shared_services_tfstate_locks"   # The DynamoDb table for the state file locking

        [[aws-account]]
        account_number = 1234
        deploy_region = 'eu-central-1'
        type = 'test'
        env_var_aws_key = 'AWS_TEST_KEY'
        env_var_aws_secret = 'AWS_TEST_SECRET'
            [aws-account.terraform.state]
          # An optional bracket. This will override the default
          # s3 bucket and path generated for maintaining state.
          s3_region = "eu-central-1"				# region of the S3 bucket
          s3_bucket = "cb_shared_services_tf_statefiles"        # S3 bucket holding the terraform state file
          s3_path = "/exampleproject/test/tf.state"             # Path in the S3 bucket where the terraform state file lives
          dynamodb_table = "cb_shared_services_tfstate_locks"   # The DynamoDb table for the state file locking

        [[aws-account]]
        account_number = 12345
        deploy_region = 'us-east-1'
        type = 'stage'
        env_var_aws_key = 'AWS_STAGE_KEY'
        env_var_aws_secret = 'AWS_STAGE_SECRET'
            [aws-account.terraform.state]
          # An optional bracket. This will override the default
          # s3 bucket and path generated for maintaining state.
          s3_region = "us-east-1"				# region of the S3 bucket
          s3_bucket = "cb_shared_services_tf_statefiles"        # S3 bucket holding the terraform state file
          s3_path = "/exampleproject/test/tf.state"             # Path in the S3 bucket where the terraform state file lives
          dynamodb_table = "cb_shared_services_tfstate_locks"   # The DynamoDb table for the state file locking

        [[aws-account]]
        account_number = 123456
        deploy_region = 'us-east-1'
        type = 'prod'
        env_var_aws_key = 'AWS_PROD_KEY'
        env_var_aws_secret = 'AWS_PROD_SECRET'
            [aws-account.terraform.state]
          # An optional bracket. This will override the default
          # s3 bucket and path generated for maintaining state.
          s3_region = "us-east-1"				# region of the S3 bucket
          s3_bucket = "cb_shared_services_tf_statefiles"        # S3 bucket holding the terraform state file
          s3_path = "/exampleproject/test/tf.state"             # Path in the S3 bucket where the terraform state file lives
          dynamodb_table = "cb_shared_services_tfstate_locks"   # The DynamoDb table for the state file locking
        """)

    # Create our mock result
    data['mock_results'] = dict()
    data['mock_results']["terraform fmt -check=true"] = Result("")
    data['mock_results']["git rev-list remotes/origin/master..remotes/origin/master --"] = \
        Result("")
    data['mock_results']["terraform validate terraform "] = Result("")
    data['mock_results']["tflint"] = Result("")

    # setup the environment:
    data['environment'] = dict(
        PD_LOW_KEY="PD_LOW_KEY",
        PD_HIGH_KEY="PD_HIGH_KEY",
        GL_PRIVATE_TOKEN="GL_PRIVATE_TOKEN",
        CI_PROJECT_PATH_SLUG="test",
        GITLAB_CI="TEST",
        CI_PROJECT_ID="TEST_Project",
        CI_COMMIT_SHA="1234",
        AWS_PROD_KEY="abcdtdg",
        AWS_PROD_SECRET="12345345",
        AWS_STAGE_KEY="efghsdfgg",
        AWS_STAGE_SECRET="567823456",
        AWS_TEST_KEY="abcdtdg",
        AWS_TEST_SECRET="12345345",
    )
    return data

def test_validate_fmt_fail(fixture_run_data):
    """Testing Terraform format failure"""
    mock_results = fixture_run_data['mock_results'].copy()
    mock_results['terraform fmt -check=true'] = Result("", stderr="/some/path was formatted", exited=1)
    mock_context = MockContext(run=mock_results)
    # Run the phase, see what happens!
    return_code = run(
        mock_context,
        branch_name="master",
        config_file=fixture_run_data['config_fd'],
        o_init_tf=False)
    assert return_code == ErrorCodes.ERROR_TERRAFORM_FORMAT

def test_validate_tfvalidate_fail(fixture_run_data):
    """Testing stub functions"""
    override_command = ""
    mock_results = fixture_run_data['mock_results'].copy()
    mock_results[f'terraform validate {override_command} terraform'] = Result("", exited=1)
    mock_context = MockContext(run=mock_results)
    return_code = run(mock_context,
                      branch_name="master",
                      config_file=fixture_run_data['config_fd'],
                      o_init_tf=False)
    assert return_code == ErrorCodes.ERROR_TERRAFORM_VALIDATE


def test_validate_tflint_fail(fixture_run_data):
    """Testing stub functions"""
    mock_results = fixture_run_data['mock_results'].copy()
    mock_results['tflint'] = Result("", exited=1)
    mock_context = MockContext(run=mock_results)
    return_code = run(
        mock_context,
        branch_name="master",
        config_file=fixture_run_data['config_fd'],
        o_init_tf=False)
    assert return_code == ErrorCodes.ERROR_TF_LINT


def test_validate_branch_diverged(fixture_run_data):
    """ Test FAILURE case. Testing stub functions"""
    mock_results = fixture_run_data['mock_results'].copy()
    mock_results['git rev-list remotes/origin/master..remotes/origin/master --'] = Result("some-commit-sha1-example")
    mock_context = MockContext(run=mock_results)
    return_code = run(
        mock_context,
        branch_name="master",
        config_file=fixture_run_data['config_fd'],
        o_init_tf=False,
        environment=fixture_run_data['environment'])

    assert return_code == ErrorCodes.ERROR_BRANCH_DIVERGED

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_validate_tfinit_fail(fixture_run_data):
    """Testing stub functions"""
    mock_results = fixture_run_data['mock_results'].copy()
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = Result("", exited=1)

    mock_context = MockContext(run=mock_results)
    return_code = run(
        mock_context,
        branch_name="master",
        config_file=fixture_run_data['config_fd'],
        environment=fixture_run_data['environment'])
    assert return_code == ErrorCodes.ERROR_TERRAFORM_INIT

@mock.patch('code.src.summoner.common.gen_backend_config_block', mock.Mock())
@mock.patch('code.src.summoner.common.backend_config_block_exists', mock.Mock())
def test_validate_success(fixture_run_data):
    """Testing stub functions"""
    override_command = ""
    mock_results = fixture_run_data['mock_results'].copy()
    mock_results['terraform init -reconfigure -backend-config=_s3_backend.summoner -input=false terraform'] = [Result("", exited=0)]*2
    mock_results[f'terraform validate {override_command} terraform'] = [Result("", exited=0)]*2

    mock_context = MockContext(run=mock_results)
    return_code = run(mock_context,
                      branch_name="master",
                      config_file=fixture_run_data['config_fd'],
                      environment=fixture_run_data['environment'])
    assert return_code == ErrorCodes.SUCCESS
